# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.testing.mocking."""

import functools
import pickle

import numpy as np
import pytest
import tensorflow as tf
# Import the final API to:
# * Register datasets
# * Make sure `tfds.load`, `tfds.builder` aliases works correctly after patching
import tensorflow_datasets as tfds


# TODO(pytest): Rather than `request.param` magic, should use
# `@pytest.mark.parametrize` once
# https://github.com/pytest-dev/pytest/issues/3960 is fixed
@pytest.fixture(
    params=[
        tfds.testing.MockPolicy.USE_FILES,
        tfds.testing.MockPolicy.USE_CODE,
    ],
)
def mock_data(request):
  """Parametrized fixture to test both `USE_FILES` and `USE_CODE` policy."""
  return functools.partial(tfds.testing.mock_data, policy=request.param)


# pylint: disable=redefined-outer-name


@pytest.fixture
def apply_mock_data(mock_data):
  """Fixture which apply `tfds.testing.mock_data` to the test.

  Test which uses this fixture will be executed twice, once with `USE_FILES` and
  once with `USE_CODE`.

  Args:
    mock_data: `mock_data` fixture defined above.

  Yields:
    None
  """
  with mock_data():
    yield


@pytest.mark.usefixtures('apply_mock_data')
def test_mocking_imagenet():
  ds = tfds.load('imagenet2012', split='train')
  assert ds.element_spec == {
      'file_name': tf.TensorSpec(shape=(), dtype=tf.string),
      'image': tf.TensorSpec(shape=(None, None, 3), dtype=tf.uint8),
      'label': tf.TensorSpec(shape=(), dtype=tf.int64),
  }
  list(ds.take(3))  # Iteration should work


@pytest.mark.usefixtures('apply_mock_data')
def test_mocking_add_tfds_id():
  read_config = tfds.ReadConfig(add_tfds_id=True)
  ds = tfds.load('mnist', split='train', read_config=read_config)
  assert ds.element_spec == {
      'tfds_id': tf.TensorSpec(shape=(), dtype=tf.string),
      'image': tf.TensorSpec(shape=(28, 28, 1), dtype=tf.uint8),
      'label': tf.TensorSpec(shape=(), dtype=tf.int64),
  }
  train_examples = list(ds.take(3))  # Iteration should work
  ds = tfds.load('mnist', split='test', read_config=read_config)
  test_examples = list(ds.take(3))  # Iteration should work
  assert train_examples[0]['tfds_id'] != test_examples[0]['tfds_id']


@pytest.mark.usefixtures('apply_mock_data')
def test_mocking_partial_decoding():
  ds = tfds.load(
      'mnist',
      split='train',
      decoders=tfds.decode.PartialDecoding({
          'image': tfds.features.Image(shape=(None, None, 1)),
      }),
  )
  assert ds.element_spec == {
      'image': tf.TensorSpec(shape=(28, 28, 1), dtype=tf.uint8),
  }
  list(ds.take(3))  # Iteration should work


@pytest.mark.usefixtures('apply_mock_data')
def test_mocking_imagenet_decoders():
  """Test with SkipDecoding."""
  ds, ds_info = tfds.load(
      'imagenet2012',
      split='train',
      decoders={'image': tfds.decode.SkipDecoding()},
      with_info=True,
  )
  assert ds.element_spec == {
      'file_name': tf.TensorSpec(shape=(), dtype=tf.string),
      'image': tf.TensorSpec(shape=(), dtype=tf.string),  # Encoded images
      'label': tf.TensorSpec(shape=(), dtype=tf.int64),
  }
  for ex in ds.take(10):
    # Image decoding should works
    image = ds_info.features['image'].decode_example(ex['image'])
    image.shape.assert_is_compatible_with((None, None, 3))
    assert image.dtype == tf.uint8


@pytest.mark.usefixtures('apply_mock_data')
def test_mocking_wider_face():
  ds = tfds.load('wider_face', split='train')
  assert ds.element_spec['faces']['expression'] == tf.TensorSpec(
      shape=(None,), dtype=tf.bool
  )
  for ex in ds.take(2):
    assert ex['faces']['expression'].dtype == tf.bool


@pytest.mark.usefixtures('apply_mock_data')
def test_mocking_coco_captions():
  ds = tfds.load('coco_captions', split='train')
  assert ds.element_spec['captions']['text'] == tf.TensorSpec(
      shape=(None,), dtype=tf.string
  )
  for ex in ds.take(2):
    assert ex['captions']['text'].dtype == tf.string
    ex['captions']['text'].shape.assert_is_compatible_with((None,))


def test_custom_as_dataset(mock_data):
  def _as_dataset(self, *args, **kwargs):  # pylint: disable=unused-argument
    return tf.data.Dataset.from_generator(
        lambda: (  # pylint: disable=g-long-lambda
            {'text': t} for t in ['some sentence', 'some other sentence']
        ),
        output_types=self.info.features.dtype,
        output_shapes=self.info.features.shape,
    )

  with mock_data(as_dataset_fn=_as_dataset):
    ds = tfds.load('librispeech_lm', split='train')
    out = [ex['text'] for ex in tfds.as_numpy(ds)]
    assert out == [b'some sentence', b'some other sentence']


def test_max_values(mock_data):
  with mock_data(num_examples=50):
    ds = tfds.load('mnist', split='train')
    assert ds.element_spec == {
        'image': tf.TensorSpec(shape=(28, 28, 1), dtype=tf.uint8),
        'label': tf.TensorSpec(shape=(), dtype=tf.int64),
    }
    for ex in ds.take(50):
      assert tf.math.reduce_max(ex['label']).numpy() < 10
    # Test determinism (iterating twice should yield the same samples)
    assert [ex['label'].numpy() for ex in ds.take(5)] == [8, 2, 9, 3, 1]
    assert [ex['label'].numpy() for ex in ds.take(5)] == [8, 2, 9, 3, 1]


def test_mock_data_auto(tmp_path):
  """Test `MockPolicy.AUTO` fallback to `USE_CODE`."""
  # By default, mock data should load metadata when present.
  with tfds.testing.mock_data():
    builder = tfds.builder('mnist')
    assert list(builder.info.splits.keys())  # Metadata should loaded.

  # When mock data unknown, fallback to `USE_CODE` mode.
  with tfds.testing.mock_data(data_dir=tmp_path):
    builder = tfds.builder('mnist')
    assert not list(builder.info.splits.keys())  # Metadata unknown.


def test_mock_data_use_code():
  """Test `MockPolicy.USE_CODE` specific behavior."""
  with tfds.testing.mock_data(policy=tfds.testing.MockPolicy.USE_CODE):
    builder = tfds.builder('mnist')
    # Dynamic metadata should be unknown.
    assert not list(builder.info.splits.keys())

    # As splits are unknown, any split can be loaded.
    ds = tfds.load('mnist', split='non_existent')
    assert set(ds.element_spec.keys()) == {'image', 'label'}

    ds = tfds.data_source('mnist', split='non_existent')
    assert len(ds) == 1
    assert set(ds[0].keys()) == {'image', 'label'}


def test_mock_data_use_files(tmp_path):
  """Test `MockPolicy.USE_FILES` specific behavior."""
  with tfds.testing.mock_data(policy=tfds.testing.MockPolicy.USE_FILES):
    builder = tfds.builder('mnist')
    # Metadata should have been restored correctly.
    assert builder.info.splits.keys() == {'test', 'train'}

    # Unknown split should raise error
    # Currently, this error is accidentally triggered by
    # `info.splits[split].file_instructions` inside `_should_cache_ds`.
    # We could make the check more explicit.
    with pytest.raises(ValueError, match='Unknown split'):
      tfds.load('mnist', split='non_existent')
    with pytest.raises(ValueError, match='Unknown split'):
      tfds.data_source('mnist', split='non_existent')

  with tfds.testing.mock_data(
      policy=tfds.testing.MockPolicy.USE_FILES,
      data_dir=tmp_path,
  ):
    with pytest.raises(ValueError, match='copy the real metadata files'):
      tfds.load('mnist')
    with pytest.raises(ValueError, match='copy the real metadata files'):
      tfds.data_source('mnist', split='non_existent')


def test_cardinality():
  with tfds.testing.mock_data(num_examples=8):
    ds = tfds.load('mnist', split='train')
    assert ds.cardinality().numpy().item() == 8

  with tfds.testing.mock_data(num_examples=15):
    ds = tfds.load('mnist', split='train')
    assert ds.cardinality().numpy().item() == 15


@pytest.mark.parametrize(
    'ds_name',
    [
        'dummy_dataset',
    ],
)
def test_mock_non_registered_datasets(
    dummy_dataset: tfds.testing.DummyDataset,
    ds_name: str,
):
  # Without mocking, the dataset cannot be found
  with pytest.raises(tfds.core.registered.DatasetNotFoundError):
    # Do not test 'huggingface:dummy_dataset' to not have tests
    # access non-hermetic resources.
    tfds.builder('dummy_dataset')

  data_dir = dummy_dataset._data_dir_root
  # After mocking, the dataset is restored from the metadata files.
  with tfds.testing.mock_data(data_dir=data_dir, num_examples=15):
    builder = tfds.builder(ds_name)
    ds = builder.as_dataset(split='train')
    assert len(list(ds)) == 15


def test_mocking_rlu_nested_dataset():
  """Test of a nested dataset.

  In this test we use the dataset rlu_atari.
  The dataset has the following features:

    features=tfds.features.FeaturesDict({
      'episode_id': tf.int64,
      'checkpoint_id': tf.int64,
      'episode_return': tf.float32,
      'steps': tfds.features.Dataset({
          'action': tf.int64,
          'discount': tf.float32,
          'is_first': tf.bool,
          'is_last': tf.bool,
          'is_terminal': tf.bool,
          'observation': tfds.features.Image(shape=(84, 84, 1), dtype=tf.uint8),
          'reward': tf.float32,
      }),
    })
  """
  with tfds.testing.mock_data(
      num_examples=3, policy=tfds.testing.MockPolicy.USE_CODE
  ):
    ds = tfds.load('rlu_atari/Pong_run_1', split='train')

    steps = ds.element_spec['steps']
    assert isinstance(steps, tf.data.DatasetSpec)
    assert steps.element_spec['reward'] == tf.TensorSpec(
        shape=(), dtype=tf.float32
    )

    for ex in ds.take(3):
      ds_steps = ex['steps']
      assert isinstance(ds_steps, tf.data.Dataset)

      ds_steps_iter = iter(ds_steps)
      steps_ex = next(ds_steps_iter)
      assert set(steps_ex.keys()) == {
          'action',
          'discount',
          'is_first',
          'is_last',
          'is_terminal',
          'observation',
          'reward',
      }
      assert steps_ex['observation'].shape == (84, 84, 1)


def _get_steps(data, window_size=4):
  """Extract the steps dataset and create out of it a window dataset."""
  episode_ds = data['steps']
  # The line below creates a variant dataset
  return episode_ds.window(window_size, drop_remainder=True)


@pytest.mark.parametrize('num_sub_examples', [1, 36])
def test_mocking_rlu_nested_dataset_with_windows(
    num_sub_examples, num_examples=3, max_value=8, window_size=4
):
  """Test of a nested dataset with windows.

  In this test we use the dataset rlu_atari - see the docstring of
  test_mocking_rlu_nested_dataset for a full list of features.

  The test checks in particular that after application of the window method
  the number of elements in the dataset is

  num_examples * (num_sub_examples // window_size).

  Args:
    num_sub_examples: Number of examples to generate in a nested subdataset.
    num_examples: Number of examples to generate in the dataset.
    max_value: The maximum value present in generated tensors.
    window_size: The size of the sequence window.
  """
  with tfds.testing.mock_data(
      num_examples=num_examples,
      num_sub_examples=num_sub_examples,
      max_value=max_value,
      policy=tfds.testing.MockPolicy.USE_CODE,
  ):
    ds = tfds.load('rlu_atari/Pong_run_1', split='train')

    for ex in ds.take(3):
      ds_steps = ex['steps']
      assert ds_steps.cardinality().numpy().item() == num_sub_examples

      # the window method is applied in _get_steps
      ds_flat_steps = ds.flat_map(
          functools.partial(_get_steps, window_size=window_size)
      )
      ds_flat_steps = iter(ds_flat_steps)

      assert len(list(ds_flat_steps)) == num_examples * (
          num_sub_examples // window_size
      )

      for obs_rew_act in ds_flat_steps:
        assert obs_rew_act['observation'].element_spec == tf.TensorSpec(
            shape=(84, 84, 1), dtype=tf.uint8
        )
        assert (
            next(iter(tfds.as_numpy(obs_rew_act['observation']))) <= max_value
        ).all()
        assert (
            next(iter(tfds.as_numpy(obs_rew_act['action']))) <= max_value
        ).all()


def test_mock_data_source():
  with tfds.testing.mock_data(num_examples=10):
    data_source = tfds.data_source('imagenet2012')
    assert len(data_source['train']) == 10
    assert isinstance(data_source['train'][0], dict)

    data_source = tfds.data_source('imagenet2012', split='train')
    assert len(data_source) == 10
    assert isinstance(data_source[0], dict)

    data_source = tfds.data_source('imagenet2012', split='train[:50%]')
    assert len(data_source) == 10
    assert isinstance(data_source[0], dict)
    assert isinstance(data_source[0]['image'], np.ndarray)

    # Without decoding the images
    decoders = {'image': tfds.decode.SkipDecoding()}
    data_source = tfds.data_source(
        'imagenet2012', split='train[:50%]', decoders=decoders
    )
    assert isinstance(data_source[0]['image'], bytes)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module including definitions for all RLU datasets."""

from __future__ import annotations

import os
from typing import Any, Dict, Generator, List, Tuple

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


_DESCRIPTION = """
RL Unplugged is suite of benchmarks for offline reinforcement learning. The RL
Unplugged is designed around the following considerations: to facilitate ease of
use, we provide the datasets with a unified API which makes it easy for the
practitioner to work with all data in the suite once a general pipeline has been
established.

The datasets follow the [RLDS format](https://github.com/google-research/rlds)
to represent steps and episodes.

"""


_HOMEPAGE = (
    'https://github.com/deepmind/deepmind-research/tree/master/rl_unplugged'
)


def filename(prefix: str, num_shards: int, shard_id: int):
  return os.fspath(
      tfds.core.Path(f'{prefix}-{shard_id:05d}-of-{num_shards:05d}')
  )


def get_files(prefix: str, num_shards: int) -> List[str]:
  return [filename(prefix, num_shards, i) for i in range(num_shards)]  # pytype: disable=bad-return-type  # gen-stub-imports


def float_tensor_feature(size: int) -> tfds.features.Tensor:
  return tfds.features.Tensor(
      shape=(size,), dtype=np.float32, encoding=tfds.features.Encoding.ZLIB
  )


class RLUBuilder(tfds.core.GeneratorBasedBuilder, skip_registration=True):
  """DatasetBuilder for RLU."""


  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION + self.get_description(),
        features=self.get_features_dict(),
        supervised_keys=None,  # disabled
        homepage=_HOMEPAGE,
        citation=self.get_citation(),
    )

  def get_file_prefix(self):
    raise NotImplementedError()

  def get_citation(self):
    raise NotImplementedError

  def get_description(self):
    raise NotImplementedError()

  def num_shards(self):
    raise NotImplementedError()

  def get_features_dict(self):
    raise NotImplementedError()

  def get_episode_id(self, episode):
    # The key of the episode is converted to string because int64 is not
    # supported as key.
    return str(episode['episode_id'])

  def tf_example_to_step_ds(
      self, tf_example: tf.train.Example
  ) -> Dict[str, Any]:
    """Create an episode from a TF example."""
    raise NotImplementedError()

  def get_splits(self):
    paths = {
        'file_paths': get_files(
            prefix=self.get_file_prefix(), num_shards=self.num_shards()
        ),
    }
    return {
        'train': self._generate_examples(paths),
    }

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    del dl_manager

    return self.get_splits()

  def generate_examples_one_file(
      self, path
  ) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
    """Yields examples from one file."""
    # Dataset of tf.Examples containing full episodes.
    example_ds = tf.data.TFRecordDataset(
        filenames=str(path), compression_type='GZIP'
    )
    # Dataset of episodes, each represented as a dataset of steps.
    episode_ds = example_ds.map(
        self.tf_example_to_step_ds,
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    episode_ds = tfds.as_numpy(episode_ds)
    for e in episode_ds:
      yield self.get_episode_id(e), e

  def _generate_examples(self, paths):
    """Yields examples."""
    beam = tfds.core.lazy_imports.apache_beam
    file_paths = paths['file_paths']

    return beam.Create(file_paths) | beam.FlatMap(
        self.generate_examples_one_file
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for XTREME-S Benchmark."""
import tensorflow as tf
from tensorflow_datasets.audio.xtreme_s import xtreme_s
import tensorflow_datasets.public_api as tfds


class FleursTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for FLEURS dataset with af_za."""

  DATASET_CLASS = xtreme_s.XtremeS
  BUILDER_CONFIG_NAMES_TO_TEST = ['fleurs.af_za']
  SKIP_CHECKSUMS = True
  SPLITS = {
      'train': 3,  # Number of fake train examples
      'validation': 1,  # Number of fake dev examples
      'test': 2,  # Number of fake test examples
  }
  DL_EXTRACT_RESULT = {'af_za': 'fleurs'}

  def _download_and_prepare_as_dataset(self, builder):
    super()._download_and_prepare_as_dataset(builder)
    if not tf.executing_eagerly():  # Only test the following in eager mode.
      return

    def get_sorted_data(builder, split):
      splits = builder.as_dataset()
      data = list(tfds.as_numpy(splits[split]))
      return sorted(data, key=lambda x: x['path'])

    # Get the first example per split, sorting by path for consistency.
    train_ex = get_sorted_data(builder, tfds.Split.TRAIN)[0]
    dev_ex = get_sorted_data(builder, tfds.Split.VALIDATION)[0]
    test_ex = get_sorted_data(builder, tfds.Split.TEST)[0]
    # Check Train
    self.assertEqual(1348, train_ex['id'])
    self.assertIn(
        b'dummy_data/fleurs/af_za/audio/train/train_1.wav', train_ex['path']
    )
    self.assertEqual('Boot ritte.'.encode(), train_ex['raw_transcription'])
    self.assertEqual('boot ritte'.encode(), train_ex['transcription'])
    self.assertEqual(32000, train_ex['num_samples'])
    self.assertEqual(1, train_ex['gender'])
    self.assertEqual(0, train_ex['lang_id'])
    self.assertEqual('Afrikaans'.encode(), train_ex['language'])
    self.assertEqual(3, train_ex['lang_group_id'])
    self.assertEqual((32000,), train_ex['audio'].shape)
    self.assertCountEqual(
        [
            'id',
            'path',
            'audio',
            'raw_transcription',
            'transcription',
            'num_samples',
            'gender',
            'lang_id',
            'language',
            'lang_group_id',
        ],
        train_ex.keys(),
    )
    # Check Dev
    self.assertEqual(305, dev_ex['id'])
    self.assertIn(
        b'dummy_data/fleurs/af_za/audio/dev/dev_1.wav', dev_ex['path']
    )
    self.assertEqual('In die 1960s'.encode(), dev_ex['raw_transcription'])
    self.assertEqual('in die 1960s'.encode(), dev_ex['transcription'])
    self.assertEqual(32000, dev_ex['num_samples'])
    self.assertEqual(0, dev_ex['gender'])
    self.assertEqual(0, dev_ex['lang_id'])
    self.assertEqual('Afrikaans'.encode(), dev_ex['language'])
    self.assertEqual(3, dev_ex['lang_group_id'])
    self.assertEqual((32000,), dev_ex['audio'].shape)
    self.assertCountEqual(
        [
            'id',
            'path',
            'audio',
            'raw_transcription',
            'transcription',
            'num_samples',
            'gender',
            'lang_id',
            'language',
            'lang_group_id',
        ],
        dev_ex.keys(),
    )
    # Check Test
    self.assertEqual(5, test_ex['id'])
    self.assertIn(
        b'dummy_data/fleurs/af_za/audio/test/test_1.wav', test_ex['path']
    )
    self.assertEqual('Boot.'.encode(), test_ex['raw_transcription'])
    self.assertEqual('boot'.encode(), test_ex['transcription'])
    self.assertEqual(32000, test_ex['num_samples'])
    self.assertEqual(1, test_ex['gender'])
    self.assertEqual(0, test_ex['lang_id'])
    self.assertEqual('Afrikaans'.encode(), dev_ex['language'])
    self.assertEqual(3, test_ex['lang_group_id'])
    self.assertEqual((32000,), test_ex['audio'].shape)
    self.assertCountEqual(
        [
            'id',
            'path',
            'audio',
            'raw_transcription',
            'transcription',
            'num_samples',
            'gender',
            'lang_id',
            'language',
            'lang_group_id',
        ],
        test_ex.keys(),
    )


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for c4_wsrs_utils."""
import re
from unittest import mock
from absl.testing import absltest
import numpy as np
from tensorflow_datasets.text.c4_wsrs import c4_wsrs_utils


class C4WsrsUtilsTest(absltest.TestCase):

  def test_create_word_finder_regex(self):
    words = ['pt', 'ct', 's']
    word_finder_re = c4_wsrs_utils.create_word_finder_regex(words)
    string = "the (pt's) #ct to feel drowsy."
    self.assertEqual(word_finder_re.findall(string), ['pt', 'ct'])

  def test_get_abbreviation_expansion_pairs(self):
    snippet = 'the patient is in the emergency room.'
    abbreviations_by_expansion = {'patient': ['pt'], 'emergency room': ['er']}
    expansion_re = re.compile('(patient|emergency room)')
    result = c4_wsrs_utils._get_abbreviation_expansion_pairs(
        snippet, abbreviations_by_expansion, expansion_re
    )
    self.assertEqual(
        result, {4: ('pt', 'patient'), 22: ('er', 'emergency room')}
    )

  def test_extract_snippets(self):
    doc = {
        'url': b'test/url.com',
        'text': b'the patient is in the emergency room. the doctor just left.',
    }
    abbreviations_by_expansion = {
        'patient': ['pt'],
        'emergency room': ['er'],
        'doctor': ['dr'],
    }
    expansion_re = re.compile('(patient|emergency room|doctor)')
    with mock.patch.object(
        np.random, 'uniform', autospec=True
    ) as random_uniform_mock:
      random_uniform_mock.return_value = 0.1
      results = list(
          c4_wsrs_utils.extract_snippets(
              doc,
              max_sentences_per_snippet=1,
              abbreviations_by_expansion=abbreviations_by_expansion,
              expansion_regex=expansion_re,
              max_snippet_char_len=1024,
              alpha_keep_no_rs=0.0,
              alpha_keep_rs=0.0,
          )
      )
    self.assertEqual(
        results,
        [
            (
                'url=test/url.com,snippet_id=0',
                (
                    'the patient is in the emergency room.',
                    ('pt', 'patient'),
                    {4: ('pt', 'patient'), 22: ('er', 'emergency room')},
                ),
            ),
            (
                'url=test/url.com,snippet_id=1',
                (
                    'the doctor just left.',
                    ('dr', 'doctor'),
                    {4: ('dr', 'doctor')},
                ),
            ),
        ],
    )

  def test_reverse_substitute_snippet(self):
    snippet = 'the patient is in the emergency room.'
    index_to_pair = {4: ('pt', 'patient'), 22: ('er', 'emergency room')}
    result = c4_wsrs_utils._reverse_substitute_snippet(
        snippet, index_to_pair, substitution_rate=1.0
    )
    self.assertEqual(result, 'the pt is in the er.')

  def test_reverse_substitution(self):
    element = (
        'url=test/url.com,snippet_id=0',
        (
            'the patient is in the emergency room.',
            ('pt', 'patient'),
            {4: ('pt', 'patient'), 22: ('er', 'emergency room')},
        ),
    )
    rs_results = list(
        c4_wsrs_utils.reverse_substitution(
            element, substitution_rate=1.0, min_snippet_token_len=3
        )
    )
    expected_features = c4_wsrs_utils.WSRSFeatures(
        original_snippet='the patient is in the emergency room.',
        abbreviated_snippet='the pt is in the er.',
    )
    self.assertSameElements(
        rs_results,
        [(
            ('pt', 'patient'),
            ('url=test/url.com,snippet_id=0', expected_features),
        )],
    )


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Winogrande Challenge."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

from etils import epath
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@article{sakaguchi2019winogrande,
    title={WinoGrande: An Adversarial Winograd Schema Challenge at Scale},
    author={Sakaguchi, Keisuke and Bras, Ronan Le and Bhagavatula, Chandra and Choi, Yejin},
    journal={arXiv preprint arXiv:1907.10641},
    year={2019}
}
"""

_DESCRIPTION = """\
The  WinoGrande, a large-scale dataset of 44k problems, inspired by the original
 Winograd Schema Challenge design, but adjusted to improve both the scale and
 the hardness of the dataset.
"""

_DATA_URL = 'https://storage.googleapis.com/ai2-mosaic/public/winogrande/winogrande_1.1.zip'


class Winogrande(tfds.core.GeneratorBasedBuilder):
  """The Winogrande challenge."""

  VERSION = tfds.core.Version('1.1.0')

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'sentence': tfds.features.Text(),
            'option1': tfds.features.Text(),
            'option2': tfds.features.Text(),
            'label': tfds.features.ClassLabel(names=['1', '2']),
        }),
        # No default supervised_keys (as we have to pass both the sentence
        # and options as input).
        supervised_keys=None,
        homepage='http://winogrande.allenai.org/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    dl_dir = dl_manager.download_and_extract(_DATA_URL)
    data_dir = os.path.join(dl_dir, 'winogrande_1.1')

    # Winogrande has different standardized training data sizes and reports
    # numbers for each of these data sizes, so make those available.
    data_sizes = ['xs', 's', 'm', 'l', 'xl']
    train_splits = []
    for size in data_sizes:
      train_splits.append(
          tfds.core.SplitGenerator(
              name=tfds.Split('train_{}'.format(size)),
              gen_kwargs={
                  'filepath': os.path.join(
                      data_dir, 'train_{}.jsonl'.format(size)
                  )
              },
          )
      )
    return train_splits + [
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={'filepath': os.path.join(data_dir, 'test.jsonl')},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={'filepath': os.path.join(data_dir, 'dev.jsonl')},
        ),
    ]

  def _generate_examples(self, filepath):
    """This function returns the examples in the raw (text) form."""
    with epath.Path(filepath).open() as f:
      for row in f:
        row_fields = json.loads(row)
        yield row_fields['qID'], {
            'sentence': row_fields['sentence'],
            'option1': row_fields['option1'],
            'option2': row_fields['option2'],
            'label': row_fields.get('answer', -1),
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for xnli dataset module."""

from tensorflow_datasets import testing
from tensorflow_datasets.text.xtreme_xnli import xtreme_xnli


class XtremeXnli(testing.DatasetBuilderTestCase):
  DATASET_CLASS = xtreme_xnli.XtremeXnli
  DL_EXTRACT_RESULT = {
      lang: f'en-{lang}.translated.tsv'
      for lang in xtreme_xnli._LANGUAGES
      if lang != 'en'
  }
  SPLITS = {'train': 2}


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset builders API."""

from tensorflow_datasets.core.dataset_builders.adhoc_builder import AdhocBuilder
from tensorflow_datasets.core.dataset_builders.adhoc_builder import store_as_tfds_dataset
from tensorflow_datasets.core.dataset_builders.adhoc_builder import TfDataBuilder
from tensorflow_datasets.core.dataset_builders.conll.conll_dataset_builder import ConllBuilderConfig
from tensorflow_datasets.core.dataset_builders.conll.conll_dataset_builder import ConllDatasetBuilder
from tensorflow_datasets.core.dataset_builders.conll.conllu_dataset_builder import ConllUBuilderConfig
from tensorflow_datasets.core.dataset_builders.conll.conllu_dataset_builder import ConllUDatasetBuilder
from tensorflow_datasets.core.dataset_builders.huggingface_dataset_builder import HuggingfaceDatasetBuilder
from tensorflow_datasets.core.dataset_builders.view_builder import ViewBuilder
from tensorflow_datasets.core.dataset_builders.view_builder import ViewConfig

__all__ = [
    "AdhocBuilder",
    "ConllBuilderConfig",
    "ConllDatasetBuilder",
    "ConllUBuilderConfig",
    "ConllUDatasetBuilder",
    "HuggingfaceDatasetBuilder",
    "store_as_tfds_dataset",
    "TfDataBuilder",
    "ViewBuilder",
    "ViewConfig",
]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.shuffle."""

import collections
import contextlib
import logging
import resource
import tempfile

from absl.testing.absltest import mock
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import shuffle
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

_ITEMS = [
    (1, b'The'),
    (2, b'quick '),
    (3, b'brown'),
    (4, b' fox '),
    (5, b'jumps'),
    ('6', b'over'),
    (b'7', b' the '),
    (8, b'lazy'),
    (9, b' dog.'),
]

_ORDERED_ITEMS_SPLIT1 = [
    b' fox ',
    b'The',
    b'over',
    b'quick ',
    b'lazy',
    b'jumps',
    b' the ',
    b' dog.',
    b'brown',
]

_ORDERED_ITEMS_SPLIT2 = [
    b' dog.',
    b'quick ',
    b'jumps',
    b' fox ',
    b' the ',
    b'brown',
    b'over',
    b'lazy',
    b'The',
]

_SORTED_ITEMS = [
    (1, b'The'),
    (2, b'quick '),
    (3, b'brown'),
    (4, b' fox '),
    (5, b'jumps'),
    (6, b'over'),
    (7, b' the '),
    (8, b'lazy'),
    (9, b' dog.'),
]

_TOTAL_SIZE = sum(len(rec) for rec in _ORDERED_ITEMS_SPLIT1)


@contextlib.contextmanager
def disable_opening_files():
  """Context manager to disable opening new files."""
  soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
  try:
    resource.setrlimit(resource.RLIMIT_NOFILE, (1, hard_limit))
    yield
  finally:
    resource.setrlimit(resource.RLIMIT_NOFILE, (soft_limit, hard_limit))


@pytest.mark.parametrize(
    [
        'num_keys',
        'num_buckets',
        'max_hkey',
        'expected_non_empty_shards',
        'expected_min_bucket_size',
        'expected_max_bucket_size',
    ],
    [
        (10, 2, 9, 2, 5, 5),
        (10, 3, 9, 3, 3, 4),
        (1024, 10, 1023, 10, 102, 103),
        (10, 2, 99, 1, 0, 10),
    ],
)
def test_get_bucket_number(
    num_keys,
    num_buckets,
    max_hkey,
    expected_non_empty_shards,
    expected_min_bucket_size,
    expected_max_bucket_size,
):
  shards = [
      shuffle.get_bucket_number(
          hkey=k, num_buckets=num_buckets, max_hkey=max_hkey
      )
      for k in range(num_keys)
  ]
  # Check shard(x) <= shard(y) if x < y.
  previous_shard = 0
  for shard in shards:
    assert shard >= previous_shard
    previous_shard = shard
  # Check distribution: all shards are used.
  counts = collections.Counter(shards)
  assert len(counts) == expected_non_empty_shards
  for bucket_size in counts.values():
    assert bucket_size >= expected_min_bucket_size
    assert bucket_size <= expected_max_bucket_size


def test_get_bucket_number_large_hkey():
  bucket = shuffle.get_bucket_number(
      hkey=314755909755515592000481005244904880883,
      num_buckets=5,
      max_hkey=314755909755515592000481005244904880883,
  )
  assert bucket == 4


def test_increase_open_files_limit(caplog):
  with disable_opening_files():
    with pytest.raises(OSError) as exc_info:
      tempfile.TemporaryFile()
    assert exc_info.value.strerror == 'Too many open files'

    shuffle._increase_open_files_limit()
    assert caplog.record_tuples == [(
        'absl',
        logging.WARNING,
        (
            'Soft limit for the maximum number of open file descriptors for the'
            ' current process increased from 1 to 1001'
        ),
    )]
    tempfile.TemporaryFile()

  _, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
  resource.setrlimit(resource.RLIMIT_NOFILE, (hard_limit, hard_limit))
  caplog.clear()
  shuffle._increase_open_files_limit()
  assert caplog.record_tuples == [(
      'absl',
      logging.ERROR,
      (
          'Soft and hard limits for the maximum number of open file descriptors'
          ' for the current process are identical.'
      ),
  )]


def test_shuffler_with_limited_open_files(tmp_path, monkeypatch, caplog):
  monkeypatch.setattr(shuffle, 'MAX_MEM_BUFFER_SIZE', 0)
  shuffler = shuffle.Shuffler(tmp_path, 'salt', disable_shuffling=False)
  # trigger Tensorflow imports before disabling opening files
  tf.io.gfile.GFile  # pylint: disable=pointless-statement
  tf.errors.ResourceExhaustedError  # pylint: disable=pointless-statement

  with disable_opening_files():
    shuffler.add(1, b'The')

  assert caplog.record_tuples == [(
      'absl',
      logging.WARNING,
      (
          'Soft limit for the maximum number of open file descriptors for the'
          ' current process increased from 1 to 1001'
      ),
  )]


class ShuffleTest(testing.TestCase):

  def _test_items(self, salt, items, expected_order, disable_shuffling=False):
    shuffler = shuffle.Shuffler(self.get_temp_dir(), salt, disable_shuffling)
    for key, item in items:
      shuffler.add(key, item)
    self.assertEqual(shuffler.size, _TOTAL_SIZE)
    if not shuffler._in_memory:  # Check size of temporary bucket files
      expected_size = (16 + 8) * len(items) + sum(len(t[1]) for t in items)
      size = 0
      for bucket in shuffler._buckets:
        if not bucket._fobj:
          continue
        bucket._fobj.close()
        with open(bucket._path, 'rb') as f:
          size += len(f.read())
      self.assertEqual(size, expected_size)
    # Check records can be read as expected:
    records = list(ex for _, ex in shuffler)
    self.assertEqual(records, expected_order)

  def test_all_mem(self):
    self._test_items('split1', _ITEMS, _ORDERED_ITEMS_SPLIT1)
    self._test_items('split2', _ITEMS, _ORDERED_ITEMS_SPLIT2)

  @mock.patch.object(shuffle, 'MAX_MEM_BUFFER_SIZE', 0)
  def test_disk(self):
    self._test_items('split1', _ITEMS, _ORDERED_ITEMS_SPLIT1)
    self._test_items('split2', _ITEMS, _ORDERED_ITEMS_SPLIT2)

  def test_sorted_by_key(self):
    self._test_items(
        'split1',
        _SORTED_ITEMS,
        [value for _, value in _SORTED_ITEMS],
        disable_shuffling=True,
    )

  def test_nonbytes(self):
    shuffler = shuffle.Shuffler(self.get_temp_dir(), 'split1')
    with self.assertRaisesWithPredicateMatch(AssertionError, 'Only bytes'):
      shuffler.add(1, 'a')
    with self.assertRaisesWithPredicateMatch(AssertionError, 'Only bytes'):
      shuffler.add(1, 123)

  def test_duplicate_key(self):
    shuffler = shuffle.Shuffler(self.get_temp_dir(), 'split1')
    shuffler.add(1, b'a')
    shuffler.add(2, b'b')
    shuffler.add(1, b'c')
    iterator = iter(shuffler)
    self.assertEqual(
        next(iterator), (86269847664267119453139349052967691808, b'a')
    )
    with self.assertRaises(shuffle.DuplicatedKeysError):
      next(iterator)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Text datasets."""

from tensorflow_datasets.text.ag_news_subset import AGNewsSubset
from tensorflow_datasets.text.anli import Anli
from tensorflow_datasets.text.assin2 import Assin2
from tensorflow_datasets.text.beir import Beir
from tensorflow_datasets.text.blimp import Blimp
from tensorflow_datasets.text.bool_q import BoolQ
from tensorflow_datasets.text.c4 import C4
from tensorflow_datasets.text.c4_wsrs import C4WSRS
from tensorflow_datasets.text.cfq import CFQ
from tensorflow_datasets.text.cfq import CFQConfig
from tensorflow_datasets.text.civil_comments import CivilComments
from tensorflow_datasets.text.clinc_oos import ClincOOS
from tensorflow_datasets.text.conll2002 import Conll2002
from tensorflow_datasets.text.conll2003 import Conll2003
from tensorflow_datasets.text.cos_e import CosE
from tensorflow_datasets.text.definite_pronoun_resolution import DefinitePronounResolution
from tensorflow_datasets.text.docnli import DocNLI
from tensorflow_datasets.text.dolphin_number_word import DolphinNumberWord
from tensorflow_datasets.text.drop import Drop
from tensorflow_datasets.text.eraser_multi_rc import EraserMultiRc
from tensorflow_datasets.text.esnli import Esnli
from tensorflow_datasets.text.gap import Gap
from tensorflow_datasets.text.gem import Gem
from tensorflow_datasets.text.glue import Glue
from tensorflow_datasets.text.goemotions import Goemotions
from tensorflow_datasets.text.gpt3 import Gpt3
from tensorflow_datasets.text.gsm8k import Gsm8k
from tensorflow_datasets.text.hellaswag import Hellaswag
from tensorflow_datasets.text.imdb import IMDBReviews
from tensorflow_datasets.text.irc_disentanglement import IrcDisentanglement
from tensorflow_datasets.text.lambada import Lambada
from tensorflow_datasets.text.librispeech_lm import LibrispeechLm
from tensorflow_datasets.text.lm1b import Lm1b
from tensorflow_datasets.text.math_dataset import MathDataset
from tensorflow_datasets.text.math_qa import MathQa
from tensorflow_datasets.text.movie_rationales import MovieRationales
from tensorflow_datasets.text.mrqa import MRQA
from tensorflow_datasets.text.multi_nli import MultiNLI
from tensorflow_datasets.text.multi_nli_mismatch import MultiNLIMismatch
from tensorflow_datasets.text.openbookqa import Openbookqa
from tensorflow_datasets.text.paws_wiki import PawsWiki
from tensorflow_datasets.text.paws_x_wiki import PawsXWiki
from tensorflow_datasets.text.pg19 import Pg19
from tensorflow_datasets.text.piqa import PIQA
from tensorflow_datasets.text.qa4mre import Qa4mre
from tensorflow_datasets.text.qrecc import QReCC
from tensorflow_datasets.text.quac import Quac
from tensorflow_datasets.text.quality import Quality
from tensorflow_datasets.text.race import Race
from tensorflow_datasets.text.reddit_disentanglement import RedditDisentanglement
from tensorflow_datasets.text.salient_span_wikipedia import SalientSpanWikipedia
from tensorflow_datasets.text.scan import Scan
from tensorflow_datasets.text.schema_guided_dialogue import SchemaGuidedDialogue
from tensorflow_datasets.text.scicite import Scicite
from tensorflow_datasets.text.scitail import SciTail
from tensorflow_datasets.text.scrolls import Scrolls
from tensorflow_datasets.text.sentiment140 import Sentiment140
from tensorflow_datasets.text.snli import Snli
from tensorflow_datasets.text.squad_question_generation import SquadQuestionGeneration
from tensorflow_datasets.text.star_cfq import StarCFQ
from tensorflow_datasets.text.story_cloze import StoryCloze
from tensorflow_datasets.text.super_glue import SuperGlue
from tensorflow_datasets.text.tiny_shakespeare import TinyShakespeare
from tensorflow_datasets.text.trec import Trec
from tensorflow_datasets.text.unifiedqa import UnifiedQA
from tensorflow_datasets.text.universal_dependencies import UniversalDependencies
from tensorflow_datasets.text.unnatural_instructions import UnnaturalInstructions
from tensorflow_datasets.text.userlibri_lm_data import UserLibriText
from tensorflow_datasets.text.wiki40b import Wiki40b
from tensorflow_datasets.text.wiki_dialog import WikiDialog
from tensorflow_datasets.text.wikiann import Wikiann
from tensorflow_datasets.text.wikipedia import Wikipedia
from tensorflow_datasets.text.wikipedia_toxicity_subtypes import WikipediaToxicitySubtypes
from tensorflow_datasets.text.winogrande import Winogrande
from tensorflow_datasets.text.wordnet import Wordnet
from tensorflow_datasets.text.wsc273 import Wsc273
from tensorflow_datasets.text.xnli import Xnli
from tensorflow_datasets.text.xtreme_pawsx import XtremePawsx
from tensorflow_datasets.text.xtreme_pos import XtremePos
from tensorflow_datasets.text.xtreme_xnli import XtremeXnli
from tensorflow_datasets.text.yelp_polarity import YelpPolarityReviews



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CivilComments from Jigsaw Unintended Bias Kaggle Competition."""

from __future__ import annotations

import ast
import csv
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

# General (main) citation for CivilComments and CivilCommentsIdentities.
_CITATION = """
@article{DBLP:journals/corr/abs-1903-04561,
  author    = {Daniel Borkan and
               Lucas Dixon and
               Jeffrey Sorensen and
               Nithum Thain and
               Lucy Vasserman},
  title     = {Nuanced Metrics for Measuring Unintended Bias with Real Data for Text
               Classification},
  journal   = {CoRR},
  volume    = {abs/1903.04561},
  year      = {2019},
  url       = {http://arxiv.org/abs/1903.04561},
  archivePrefix = {arXiv},
  eprint    = {1903.04561},
  timestamp = {Sun, 31 Mar 2019 19:01:24 +0200},
  biburl    = {https://dblp.org/rec/bib/journals/corr/abs-1903-04561},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}
"""

# Citation for CivilCommentsCovert.
_COVERT_CITATION = """
@inproceedings{lees-etal-2021-capturing,
    title = "Capturing Covertly Toxic Speech via Crowdsourcing",
    author = "Lees, Alyssa  and
      Borkan, Daniel  and
      Kivlichan, Ian  and
      Nario, Jorge  and
      Goyal, Tesh",
    booktitle = "Proceedings of the First Workshop on Bridging Human{--}Computer Interaction and Natural Language Processing",
    month = apr,
    year = "2021",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/2021.hcinlp-1.3",
    pages = "14--20"
}
"""

# Citation for CivilComments Toxic Spans.
_SPANS_CITATION = """
@inproceedings{pavlopoulos-etal-2021-semeval,
    title = "{S}em{E}val-2021 Task 5: Toxic Spans Detection",
    author = "Pavlopoulos, John  and Sorensen, Jeffrey  and Laugier, L{\'e}o and Androutsopoulos, Ion",
    booktitle = "Proceedings of the 15th International Workshop on Semantic Evaluation (SemEval-2021)",
    month = aug,
    year = "2021",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2021.semeval-1.6",
    doi = "10.18653/v1/2021.semeval-1.6",
    pages = "59--69",
}
"""

# Citation for CivilComments Context.
_CONTEXT_CITATION = """
@misc{pavlopoulos2020toxicity,
    title={Toxicity Detection: Does Context Really Matter?},
    author={John Pavlopoulos and Jeffrey Sorensen and Lucas Dixon and Nithum Thain and Ion Androutsopoulos},
    year={2020}, eprint={2006.00998}, archivePrefix={arXiv}, primaryClass={cs.CL}
}
"""

_COMMON_DESCRIPTION = """
This version of the CivilComments Dataset provides access to the primary
seven labels that were annotated by crowd workers, the toxicity and other
tags are a value between 0 and 1 indicating the fraction of annotators that
assigned these attributes to the comment text.

The other tags are only available for a fraction of the input examples. They
are currently ignored for the main dataset; the CivilCommentsIdentities set
includes those labels, but only consists of the subset of the data with them.
The other attributes that were part of the original CivilComments release are
included only in the raw data. See the Kaggle documentation for more details
about the available features.

The comments in this dataset come from an archive of the Civil Comments
platform, a commenting plugin for independent news sites. These public comments
were created from 2015 - 2017 and appeared on approximately 50 English-language
news sites across the world. When Civil Comments shut down in 2017, they chose
to make the public comments available in a lasting open archive to enable future
research. The original data, published on figshare, includes the public comment
text, some associated metadata such as article IDs, publication IDs, timestamps
and commenter-generated "civility" labels, but does not include user ids. Jigsaw
extended this dataset by adding additional labels for toxicity, identity
mentions, as well as covert offensiveness. This data set is an exact replica of
the data released for the Jigsaw Unintended Bias in Toxicity Classification
Kaggle challenge. This dataset is released under CC0, as is the underlying
comment text.

For comments that have a parent_id also in the civil comments data, the
text of the previous comment is provided as the "parent_text" feature. Note
that the splits were made without regard to this information, so using previous
comments may leak some information. The annotators did not have access to the
parent text when making the labels.
"""

_CC_DESCRIPTION = """

The CivilComments set here includes all the data, but only the basic seven
labels (toxicity, severe_toxicity, obscene, threat, insult, identity_attack, and
sexual_explicit).
"""

_CC_IDENTITIES_DESCRIPTION = """
The CivilCommentsIdentities set here includes an extended set of identity labels
in addition to the basic seven labels. However, it only includes the subset
(roughly a quarter) of the data with all these features.
"""

_CC_COVERT_DESCRIPTION = """
WARNING: there's a potential data quality issue with CivilCommentsCovert that
we're actively working on fixing (06/28/22); the underlying data may change!

The CivilCommentsCovert set is a subset of CivilCommentsIdentities with ~20% of
the train and test splits further annotated for covert offensiveness, in
addition to the toxicity and identity labels. Raters were asked to categorize
comments as one of explicitly, implicitly, not, or not sure if offensive, as
well as whether it contained different types of covert offensiveness. The full
annotation procedure is detailed in a forthcoming paper at
https://sites.google.com/corp/view/hciandnlp/accepted-papers.
"""

_CC_SPANS_DESCRIPTION = """
The CivilComments Toxic Spans are a subset of CivilComments that is
labeled at the span level - the indices of all character (unicode codepoints)
boundaries that were tagged as toxic by a majority of the annotators is
returned in a 'spans' feature.
"""

_CC_CONTEXT_DESCRIPTION = """
The CivilComments in Context is a subset of CivilComments that was
labeled by making available to the labelers the parent_text. It includes
a contextual_toxicity feature.
"""

_DOWNLOAD_URL = 'https://storage.googleapis.com/jigsaw-unintended-bias-in-toxicity-classification/civil_comments_v1.2.zip'

IDENTITY_LABELS = [
    'male',
    'female',
    'transgender',
    'other_gender',
    'heterosexual',
    'homosexual_gay_or_lesbian',
    'bisexual',
    'other_sexual_orientation',
    'christian',
    'jewish',
    'muslim',
    'hindu',
    'buddhist',
    'atheist',
    'other_religion',
    'black',
    'white',
    'asian',
    'latino',
    'other_race_or_ethnicity',
    'physical_disability',
    'intellectual_or_learning_disability',
    'psychiatric_or_mental_illness',
    'other_disability',
]

COVERT_LABELS = [
    'explicitly_offensive',
    'implicitly_offensive',
    'not_sure_offensive',
    'not_offensive',
    'covert_humor',
    'covert_obfuscation',
    'covert_emoticons_emojis',
    'covert_sarcasm',
    'covert_microaggression',
    'covert_masked_harm',
    'covert_political',
]


def _labels(mode):
  """Return the list of label features appropriate for the mode."""
  if mode == 'spans':
    return ['spans']
  labels = [
      'toxicity',
      'severe_toxicity',
      'obscene',
      'threat',
      'insult',
      'identity_attack',
      'sexual_explicit',
  ]
  if mode in ['identity', 'covert']:
    labels += IDENTITY_LABELS
  if mode == 'covert':
    labels += COVERT_LABELS
  if mode == 'context':
    labels += ['contextual_toxicity']
  return labels


def _parse_common(row):
  """Parse common elements to TF Example from CSV row for non-spans mode."""
  example = {}
  example['id'] = row['id']
  example['text'] = row['comment_text']
  example['parent_text'] = row['parent_text']
  example['publication_id'] = row['publication_id']
  example['created_date'] = row['created_date']
  parent_id = row['parent_id'] or 0
  example['parent_id'] = int(float(parent_id))
  example['article_id'] = int(row['article_id'])
  return example


def _parse_row_as_example(row, mode):
  """Parse elements to TF Example, as appropriate for specified mode."""
  example = _parse_common(row)
  for label in _labels(mode):
    if not row[label]:
      return
    example[label] = float(row[label])
  return example


def _parse_spans_row_as_example(row):
  """Parse elements to TF Example for toxic spans mode."""
  example = _parse_common(row)
  example['spans'] = np.array(ast.literal_eval(row['spans']), dtype=np.int32)
  return example


class CivilCommentsConfig(tfds.core.BuilderConfig):
  """Configuration for `CivilComments`."""

  def __init__(self, name, description, mode):
    super(CivilCommentsConfig, self).__init__(
        name=name, description=description
    )
    self.mode = mode


class CivilComments(tfds.core.GeneratorBasedBuilder):
  """Classification and tagging of 2M comments on news sites.

  This version of the CivilComments Dataset provides access to the primary
  seven labels that were annotated by crowd workers, the toxicity and other
  tags are a value between 0 and 1 indicating the fraction of annotators that
  assigned these attributes to the comment text.

  The other tags are only available for a fraction of the input examples. They
  are currently ignored for the main dataset; the CivilCommentsIdentities set
  includes those labels, but only consists of the subset of the data with them.
  The other attributes that were part of the original CivilComments release are
  included only in the raw data. See the Kaggle documentation for more details
  about the available features.
  """

  BUILDER_CONFIGS = [
      CivilCommentsConfig(
          name='CivilComments', description=_CC_DESCRIPTION, mode='base'
      ),
      CivilCommentsConfig(
          name='CivilCommentsIdentities',
          description=_CC_IDENTITIES_DESCRIPTION,
          mode='identity',
      ),
      CivilCommentsConfig(
          name='CivilCommentsCovert',
          description=_CC_COVERT_DESCRIPTION,
          mode='covert',
      ),
      CivilCommentsConfig(
          name='CivilCommentsToxicSpans',
          description=_CC_SPANS_DESCRIPTION,
          mode='spans',
      ),
      CivilCommentsConfig(
          name='CivilCommentsInContext',
          description=_CC_CONTEXT_DESCRIPTION,
          mode='context',
      ),
  ]

  VERSION = tfds.core.Version('1.2.4')
  RELEASE_NOTES = {
      '1.2.4': 'Add publication IDs and comment timestamps.',
      '1.2.3': 'Add warning to CivilCommentsCovert as we fix a data issue.',
      '1.2.2': 'Update to reflect context only having a train split.',
      '1.2.1': 'Fix incorrect formatting in context splits.',
      '1.2.0': 'Add toxic spans, context, and parent comment text features.',
      '1.1.3': 'Corrected id types from float to string.',
      '1.1.2': 'Added separate citation for CivilCommentsCovert dataset.',
      '1.1.1': 'Added CivilCommentsCovert config with correct checksum.',
      '1.1.0': 'Added CivilCommentsCovert config.',
      '1.0.1': 'Added a unique id for each comment.',
      '1.0.0': 'Initial full release.',
  }

  def _info(self):
    mode = self.builder_config.mode
    citation = {
        'base': _CITATION,
        'identity': _CITATION,
        'covert': _COVERT_CITATION,
        'spans': _SPANS_CITATION,
        'context': _CONTEXT_CITATION,
    }[mode]
    features = {
        'text': tfds.features.Text(),
        'id': np.str_,
        'parent_text': tfds.features.Text(),
        'parent_id': np.int32,
        'article_id': np.int32,
        'publication_id': np.str_,
        'created_date': np.str_,
    }
    if mode == 'spans':
      features['spans'] = tfds.features.Tensor(shape=(None,), dtype=np.int32)
      supervised_value = 'spans'
    else:
      for label in _labels(mode):
        features[label] = np.float32
      supervised_value = 'toxicity'

    return tfds.core.DatasetInfo(
        builder=self,
        description=_COMMON_DESCRIPTION,
        features=tfds.features.FeaturesDict(features),
        # The supervised_keys version is very impoverished.
        supervised_keys=('text', supervised_value),
        homepage='https://www.kaggle.com/c/jigsaw-unintended-bias-in-toxicity-classification/data',
        citation=citation,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_path = dl_manager.download_and_extract(_DOWNLOAD_URL)
    mode = self.builder_config.mode
    filename = os.path.join(dl_path, 'civil_comments.csv')

    splits = [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'filename': filename,
                'mode': mode,
                'split': 'train',
            },
        )
    ]

    if mode != 'covert' and mode != 'context':
      # Covert and context do not have validation split.
      validation_split_generator = tfds.core.SplitGenerator(
          name=tfds.Split.VALIDATION,
          gen_kwargs={
              'filename': filename,
              'mode': mode,
              'split': 'test_public',
          },
      )
      splits.append(validation_split_generator)

    if mode != 'context':
      # Context has only a train split.
      test_split_generator = tfds.core.SplitGenerator(
          name=tfds.Split.TEST,
          gen_kwargs={
              'filename': filename,
              'mode': mode,
              'split': 'test_private',
          },
      )
      splits.append(test_split_generator)
    return splits

  def _generate_examples(self, filename, mode, split):
    """Yields examples.

    Each example contains a text input followed by several toxicity subtype
    scores, identity labels if include_identity_labels is True, and covert
    offensiveness labels if include_covert_labels is True.

    Args:
      filename: the path of the file to be read for this split.
      mode: the specifc data subset and features to yield.
      split: the split to extract, one of train, public_test, private_test.

    Yields:
      A dictionary of features, depending upon the mode.
    """
    with epath.Path(filename).open() as f:
      reader = csv.DictReader(f)
      for row in reader:
        if mode == 'spans':
          if row['spans_split'] == split:
            example = _parse_spans_row_as_example(row)
            yield row['id'], example
        else:
          if row['split'] == split:
            example = _parse_row_as_example(row, mode)
            if example:
              yield row['id'], example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Image feature."""

from __future__ import annotations

import dataclasses
import functools
import io
import os
import tempfile
from typing import Any, List, Optional, Union, Type

from absl import logging
from etils import epath
import numpy as np
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.core.utils import dtype_utils
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

try:
  PIL_Image = lazy_imports_lib.lazy_imports.PIL_Image  # pylint: disable=invalid-name
except ImportError:
  PIL_Image = None  # pylint: disable=invalid-name

try:
  cv2 = lazy_imports_lib.lazy_imports.cv2
except ImportError:
  cv2 = None

Json = type_utils.Json
PilImage = Any  # Require lazy deps.


_ENCODE_FN = {
    'png': lambda: tf.image.encode_png,
    'jpeg': lambda: tf.image.encode_jpeg,
}


_ACCEPTABLE_CHANNELS = {
    'png': (0, 1, 2, 3, 4),
    'jpeg': (0, 1, 3),
}


_ACCEPTABLE_DTYPES = {
    'png': [np.uint8, np.uint16, np.float32],
    'jpeg': [np.uint8],
    None: [np.uint8, np.uint16, np.float32],
}


THUMBNAIL_SIZE = 128

# Framerate for the `tfds.as_dataframe` visualization
# Could add a framerate kwargs in __init__ to allow datasets to customize
# the output.
_VISU_FRAMERATE = 10


@dataclasses.dataclass
class _ImageEncoder:
  """Utils which encode/decode images."""

  shape: utils.Shape
  dtype: type_utils.TfdsDType
  encoding_format: Optional[str]
  np_dtype: Optional[np.dtype] = None

  def __post_init__(self):
    self.np_dtype = dtype_utils.cast_to_numpy(self.dtype)

  # TODO(tfds): Should deprecate the TFGraph runner in favor of simpler
  # implementation
  @functools.cached_property
  def _runner(self) -> utils.TFGraphRunner:
    return utils.TFGraphRunner()

  def encode_image_or_path(self, image_or_path_or_fobj):
    """Convert the given image into a dict convertible to tf example."""
    if isinstance(image_or_path_or_fobj, np.ndarray):
      encoded_image = self._encode_image(image_or_path_or_fobj)
    elif isinstance(image_or_path_or_fobj, epath.PathLikeCls):
      image_or_path_or_fobj = os.fspath(image_or_path_or_fobj)
      with tf.io.gfile.GFile(image_or_path_or_fobj, 'rb') as image_f:
        encoded_image = image_f.read()
    elif isinstance(image_or_path_or_fobj, bytes):
      encoded_image = image_or_path_or_fobj
    else:
      encoded_image = image_or_path_or_fobj.read()
    # If encoding is explicitly set, should verify that bytes match encoding.
    return encoded_image

  def _encode_image(self, np_image: np.ndarray) -> bytes:
    """Returns np_image encoded as jpeg or png."""
    _validate_np_array(np_image, shape=self.shape, dtype=self.np_dtype)

    # When encoding isn't defined, default to PNG.
    # Should we be more strict about explicitly define the encoding (raise
    # error / warning instead) ?
    # It has created subtle issues for imagenet_corrupted: images are read as
    # JPEG images to apply some processing, but final image saved as PNG
    # (default) rather than JPEG.
    return self._runner.run(
        _ENCODE_FN[self.encoding_format or 'png'](), np_image
    )

  def decode_image(self, img: tf.Tensor) -> tf.Tensor:
    """Decode the jpeg or png bytes to 3d tensor."""
    tf_dtype = tf.dtypes.as_dtype(self.dtype)
    img = tf.image.decode_image(img, channels=self.shape[-1], dtype=tf_dtype)
    img.set_shape(self.shape)
    return img


class _FloatImageEncoder(_ImageEncoder):
  """Image encoder of float32.

  Wrapper around `_ImageEncoder` which bitcast 1-channel float32 image
  into 4-channels uint8, so image can be encoded to PNG.
  """

  def __init__(
      self,
      *,
      shape: utils.Shape,
      encoding_format: str,
  ):
    # Assert that shape and encoding are valid when dtype==tf.float32.
    if shape[-1] != 1:
      raise ValueError(
          'tfds.features.Image only support single-channel for tf.float32. '
          f'Got shape={shape}'
      )
    if encoding_format and encoding_format != 'png':
      raise ValueError(
          'tfds.features.Image only support PNG encoding for tf.float32'
      )
    self._float_shape = shape
    super().__init__(
        shape=shape[:2] + (4,),
        dtype=np.uint8,
        encoding_format=encoding_format,
    )

  def encode_image_or_path(self, image_or_path_or_fobj):
    """Convert the given image into a dict convertible to tf example."""
    if not isinstance(image_or_path_or_fobj, np.ndarray):
      raise ValueError(
          'tfds.features.Image only support `np.ndarray` for tf.float32 '
          f'images, not paths. Got: {image_or_path_or_fobj!r}'
      )
    return self._encode_image(image_or_path_or_fobj)

  def _encode_image(self, np_image: np.ndarray) -> bytes:
    _validate_np_array(np_image, shape=self._float_shape, dtype=np.float32)
    # Bitcast 1 channel float32 -> 4 channels uint8
    np_image = np_image.view(np.uint8)
    np_image = super()._encode_image(np_image)
    return np_image

  def decode_image(self, img: tf.Tensor) -> tf.Tensor:
    img = super().decode_image(img)
    # Bitcast 4 channels uint8 -> 1 channel float32
    img = tf.bitcast(img, tf.float32)[..., None]
    return img


class Image(feature_lib.FeatureConnector):
  """`FeatureConnector` for images.

  During `_generate_examples`, the feature connector accept as input any of:

    * `str`: path to a {bmp,gif,jpeg,png} image (ex: `/path/to/img.png`).
    * `np.array`: 3d `np.uint8` array representing an image.
    * A file object containing the png or jpeg encoded image string (ex:
      `io.BytesIO(encoded_img_bytes)`)

  Output:

    `tf.Tensor` of type `tf.uint8` and shape `[height, width, num_channels]`
    for BMP, JPEG, and PNG images and shape `[num_frames, height, width, 3]` for
    GIF images.

  Example:

    * In the `tfds.core.DatasetInfo` object:

    ```python
    features=features.FeaturesDict({
        'input': features.Image(),
        'target': features.Image(shape=(None, None, 1), encoding_format='png'),
    })
    ```

    * During generation:

    ```python
    yield {
        'input': 'path/to/img.jpg',
        'target': np.ones(shape=(64, 64, 1), dtype=np.uint8),
    }
    ```
  """

  # If updating the signature here, LabeledImage and Video should likely be
  # updated too.
  def __init__(
      self,
      *,
      shape: Optional[utils.Shape] = None,
      dtype: Optional[type_utils.TfdsDType] = None,
      encoding_format: Optional[str] = None,
      use_colormap: bool = False,
      doc: feature_lib.DocArg = None,
  ):
    """Construct the connector.

    Args:
      shape: tuple of ints or None, the shape of decoded image. For GIF images:
        (num_frames, height, width, channels=3). num_frames, height and width
        can be None. For other images: (height, width, channels). height and
        width can be None. See `tf.image.encode_*` for doc on channels
        parameter. Defaults to (None, None, 3).
      dtype: `np.uint8` (default), `np.uint16` or `np.float32`. * `np.uint16`
        requires png encoding_format. * `np.float32` only supports
        single-channel image. Internally float images are bitcasted to
        4-channels `np.uint8` and saved as PNG.
      encoding_format: 'jpeg' or 'png'. Format to serialize `np.ndarray` images
        on disk. If None, encode images as PNG. If image is loaded from
        {bmg,gif,jpeg,png} file, this parameter is ignored, and file original
        encoding is used.
      use_colormap: Only used for gray-scale images. If `True`,
        `tfds.as_dataframe` will display each value in the image with a
        different color.
      doc: Documentation of this feature (e.g. description).

    Raises:
      ValueError: If the shape is invalid
    """
    super().__init__(doc=doc)
    # Set and validate values
    shape = shape or (None, None, 3)
    dtype = dtype_utils.cast_to_numpy(dtype or np.uint8)
    self._encoding_format = get_and_validate_encoding(encoding_format)
    self._shape = get_and_validate_shape(shape, self._encoding_format)
    self._dtype = get_and_validate_dtype(dtype, self._encoding_format)
    self._use_colormap = _get_and_validate_colormap(
        use_colormap, self._shape, self._dtype, self._encoding_format
    )

    if self._dtype == np.float32:  # Float images encoded as 4-channels uint8
      self._image_encoder = _FloatImageEncoder(
          shape=self._shape,
          encoding_format=self._encoding_format,
      )
    else:
      self._image_encoder = _ImageEncoder(
          shape=self._shape,
          dtype=self._dtype,
          encoding_format=self._encoding_format,
      )

  @property
  def encoding_format(self) -> Optional[str]:
    return self._encoding_format

  @property
  def use_colormap(self) -> bool:
    return self._use_colormap

  @py_utils.memoize()
  def get_tensor_info(self):
    # Image is returned as a 3-d uint8 tf.Tensor.
    return feature_lib.TensorInfo(shape=self._shape, dtype=self._dtype)

  @py_utils.memoize()
  def get_serialized_info(self):
    # Only store raw image (includes size).
    return feature_lib.TensorInfo(shape=(), dtype=np.object_)

  def encode_example(self, image_or_path_or_fobj):
    """Convert the given image into a dict convertible to tf example."""
    return self._image_encoder.encode_image_or_path(image_or_path_or_fobj)

  def decode_example(self, example):
    """Reconstruct the image with TensorFlow from the tf example."""
    return self._image_encoder.decode_image(example)

  def decode_example_np(self, example: bytes) -> np.ndarray:
    """Reconstruct the image with OpenCV from bytes, or default to PIL."""
    num_channels = self._shape[-1]
    if cv2 is not None:
      return self.decode_example_np_with_opencv(example, num_channels)
    elif PIL_Image is not None:
      logging.log_first_n(
          logging.WARNING,
          (
              'OpenCV is not installed. We recommend using OpenCV because it is'
              ' faster according to our benchmarks. Defaulting to PIL to decode'
              ' images...'
          ),
          1,
      )
      return self.decode_example_np_with_pil(example, num_channels)
    else:
      raise ImportError(
          'Decoding images with NumPy requires either OpenCV or PIL.\nWe'
          ' recommend using OpenCV because it is faster according to our'
          ' benchmarks.\nInstall them with: `pip install opencv-python` or `pip'
          ' install pillow`.'
      )

  def decode_example_np_with_opencv(
      self, example: bytes, num_channels: int
  ) -> np.ndarray:
    """Reconstruct the image with OpenCV from bytes."""
    assert cv2, 'OpenCV is not installed. OpenCV is required for this method.'
    example = np.frombuffer(example, dtype=np.uint8)
    example = cv2.imdecode(example, cv2.IMREAD_UNCHANGED)
    dtype = self.np_dtype if self.np_dtype != np.float32 else np.uint8
    example = example.astype(dtype, copy=False)
    example = _reorder_opencv_channels(example)
    example = _reshape_grayscale_image(example, num_channels)
    # Bitcast 4 channels uint8 -> 1 channel float32.
    if self.np_dtype == np.float32:
      return example.view(np.float32)
    return example

  def decode_example_np_with_pil(
      self, example: bytes, num_channels: int
  ) -> np.ndarray:
    assert PIL_Image, 'PIL is not installed. PIL is required for this method.'
    bytes_io = io.BytesIO(example)
    with PIL_Image.open(bytes_io) as image:
      dtype = self.np_dtype if self.np_dtype != np.float32 else np.uint8
      image = np.asarray(image, dtype=dtype)
      image = _reshape_grayscale_image(image, num_channels)
      if self.np_dtype == np.uint8:
        return image
      # Bitcast 4 channels uint8 -> 1 channel float32.
      if self.np_dtype == np.float32:
        return image.view(np.float32)
      raise ValueError(
          f'PIL does not handle {self.np_dtype} images. Please, install OpenCV'
          ' instead.'
      )

  def repr_html(self, ex: np.ndarray) -> str:
    """Images are displayed as thumbnail."""
    # Normalize image and resize
    img = utils.create_thumbnail(ex, use_colormap=self._use_colormap)

    # Convert to base64
    img_str = utils.get_base64(lambda buff: img.save(buff, format='PNG'))

    # Display HTML
    return f'<img src="data:image/png;base64,{img_str}" alt="Img" />'

  def repr_html_batch(self, ex: np.ndarray) -> str:
    """`Sequence(Image())` are displayed as `<video>`."""
    if ex.shape[0] == 1:
      ex = ex.squeeze(axis=0)  # (1, h, w, c) -> (h, w, c)
      return self.repr_html(ex)
    else:
      return make_video_repr_html(ex, use_colormap=self._use_colormap)

  @classmethod
  def from_json_content(
      cls, value: Union[Json, feature_pb2.ImageFeature]
  ) -> 'Image':
    if isinstance(value, dict):
      # For backwards compatibility
      return cls(
          shape=tuple(value['shape']),
          dtype=feature_lib.dtype_from_str(value['dtype']),
          encoding_format=value['encoding_format'],
          use_colormap=value.get('use_colormap'),
      )
    return cls(
        shape=feature_lib.from_shape_proto(value.shape),
        dtype=feature_lib.dtype_from_str(value.dtype),
        encoding_format=value.encoding_format or None,
        use_colormap=value.use_colormap,
    )

  def to_json_content(self) -> feature_pb2.ImageFeature:
    return feature_pb2.ImageFeature(
        shape=feature_lib.to_shape_proto(self._shape),
        dtype=feature_lib.dtype_to_str(self._dtype),
        encoding_format=self._encoding_format,
        use_colormap=self._use_colormap,
    )


# Visualization Video


def make_video_repr_html(ex, *, use_colormap: bool):
  """Returns the encoded `<video>` or GIF <img/> HTML."""
  # Use GIF to generate a HTML5 compatible video if FFMPEG is not
  # installed on the system.
  images = [
      utils.create_thumbnail(frame, use_colormap=use_colormap) for frame in ex
  ]

  if not images:
    return 'Video with 0 frames.'

  # Display the video HTML (either GIF of mp4 if ffmpeg is installed)
  try:
    utils.ffmpeg_run(['-version'])  # Check for ffmpeg installation.
  except FileNotFoundError:
    # print as `stderr` is displayed poorly on Colab
    print('FFMPEG not detected. Falling back on GIF.')
    return _get_repr_html_gif(images)
  else:
    return _get_repr_html_ffmpeg(images)


def _get_repr_html_ffmpeg(images: List[PilImage]) -> str:
  """Runs ffmpeg to get the mp4 encoded <video> str."""
  # Find number of digits in len to give names.
  num_digits = len(str(len(images))) + 1
  with tempfile.TemporaryDirectory() as video_dir:
    for i, img in enumerate(images):
      f = os.path.join(video_dir, f'img{i:0{num_digits}d}.png')
      img.save(f, format='png')

    ffmpeg_args = [
        '-framerate',
        str(_VISU_FRAMERATE),
        '-i',
        os.path.join(video_dir, f'img%0{num_digits}d.png'),
        # Using native h264 to encode video stream to H.264 codec
        # Default encoding does not seems to be supported by chrome.
        '-vcodec',
        'h264',
        # When outputting H.264, `-pix_fmt yuv420p` maximize compatibility
        # with bad video players.
        # Ref: https://trac.ffmpeg.org/wiki/Slideshow
        '-pix_fmt',
        'yuv420p',
        # ffmpeg require height/width to be even, so we rescale it
        # https://stackoverflow.com/questions/20847674/ffmpeg-libx264-height-not-divisible-by-2
        '-vf',
        'pad=ceil(iw/2)*2:ceil(ih/2)*2',
        # Native encoder cannot encode images of small scale
        # or the the hardware encoder may be busy which raises
        # Error: cannot create compression session
        # so allow software encoding
        # '-allow_sw', '1',
    ]
    video_path = epath.Path(video_dir) / 'output.mp4'
    ffmpeg_args.append(os.fspath(video_path))
    utils.ffmpeg_run(ffmpeg_args)
    video_str = utils.get_base64(video_path.read_bytes())
  return (
      f'<video height="{THUMBNAIL_SIZE}" width="175" '
      'controls loop autoplay muted playsinline>'
      f'<source src="data:video/mp4;base64,{video_str}"  type="video/mp4" >'
      '</video>'
  )


def _get_repr_html_gif(images: List[PilImage]) -> str:
  """Get the <img/> str."""

  def write_buff(buff):
    images[0].save(
        buff,
        format='GIF',
        save_all=True,
        append_images=images[1:],
        duration=1000 / _VISU_FRAMERATE,
        loop=0,
    )

  # Convert to base64
  gif_str = utils.get_base64(write_buff)
  return f'<img src="data:image/png;base64,{gif_str}" alt="Gif" />'


# Other image utils


def get_and_validate_encoding(encoding_format: Optional[str]) -> Optional[str]:
  """Update the encoding format."""
  supported = _ENCODE_FN.keys()
  if encoding_format and encoding_format not in supported:
    raise ValueError(f'`encoding_format` must be one of {supported}.')
  return encoding_format


def get_and_validate_dtype(dtype: np.dtype, encoding_format: Optional[str]):
  """Update the dtype."""
  acceptable_dtypes = _ACCEPTABLE_DTYPES.get(encoding_format)
  if acceptable_dtypes and dtype not in acceptable_dtypes:
    raise ValueError(
        f'Acceptable `dtype` for {encoding_format}: '
        f'{acceptable_dtypes} (was {dtype})'
    )
  return dtype


def get_and_validate_shape(shape, encoding_format):
  """Update the shape."""
  if len(shape) <= 2:
    raise ValueError(f'Image shape should be (h, w, c). Got: {shape}')
  channels = shape[-1]
  acceptable_channels = _ACCEPTABLE_CHANNELS.get(encoding_format)
  if acceptable_channels and channels not in acceptable_channels:
    raise ValueError(
        f'Acceptable `channels` for {encoding_format}: '
        f'{acceptable_channels} (was {channels})'
    )
  return tuple(shape)


def _get_and_validate_colormap(
    use_colormap, shape, dtype: np.dtype, encoding_format
):
  """Validate that the given colormap is valid."""
  if use_colormap:
    if encoding_format and encoding_format != 'png':
      raise ValueError(
          f'Colormap is only available for PNG images. Got: {encoding_format}'
      )
    if shape[-1] != 1:
      raise ValueError(
          f'Colormap is only available for gray-scale images. Got: {shape}'
      )
    if not dtype_utils.is_integer(dtype):
      raise ValueError(
          f'Colormap is only available for integer images. Got: {dtype}'
      )

  return use_colormap


def _validate_np_array(
    np_array: np.ndarray,
    shape: utils.Shape,
    dtype: Union[np.dtype, Type[np.generic]],
) -> None:
  """Validate the numpy array match the expected shape/dtype."""
  if np_array.dtype != dtype:
    raise ValueError(
        f'Image dtype should be {dtype}. Detected: {np_array.dtype}.'
    )
  utils.assert_shape_match(np_array.shape, shape)


def _reorder_opencv_channels(example: np.ndarray) -> np.ndarray:
  """Restore channels in the expected order: OpenCV uses BGR rather than RGB."""
  if example.ndim == 2:
    return example
  assert cv2, 'OpenCV is not installed. OpenCV is required for this method.'
  num_channels = example.shape[-1]
  if num_channels == 3:
    return cv2.cvtColor(example, cv2.COLOR_BGR2RGB)
  elif num_channels == 4:
    return cv2.cvtColor(example, cv2.COLOR_BGRA2RGBA)
  else:
    return example


def _reshape_grayscale_image(
    image: np.ndarray, num_channels: int
) -> np.ndarray:
  """Reshape grayscale images: (h, w) or (h, w, 1) -> (h, w, num_channels).

  This reproduces the transformation TensorFlow applies to grayscale images.

  Args:
    image: An image as an np.ndarray.
    num_channels: The number of channels in the image feature.

  Returns:
    The reshaped image.
  """
  if image.ndim == 2:  # (h, w)
    return np.repeat(image[..., None], num_channels, axis=-1)
  if image.ndim == 3 and image.shape[2] == 1:  # (h, w, 1)
    return np.repeat(image, num_channels, axis=-1)
  return image



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.utils.benchmark."""

import pytest

import tensorflow as tf
from tensorflow_datasets.core.utils import benchmark


@pytest.mark.parametrize(
    'ds, num_iter, batch_size, expected_num_examples',
    [
        (tf.data.Dataset.range(10), 10, 1, 10),
        (tf.data.Dataset.range(10), 5, 1, 5),
        (tf.data.Dataset.range(10), 10, 2, 20),
        (range(10), 10, 1, 10),
        (range(10), 10, 2, 20),
    ],
)
def test_raw_benchmark(ds, num_iter, batch_size, expected_num_examples):
  result = benchmark.raw_benchmark(ds, num_iter=num_iter, batch_size=batch_size)
  assert isinstance(result, benchmark.RawBenchmarkResult)
  assert result.num_examples == expected_num_examples
  assert result.end_time > result.start_time
  assert result.end_time >= result.first_batch_time


def test_benchmark():
  # Works with tf.data.Dataset
  ds = tf.data.Dataset.range(10)
  result = benchmark.benchmark(ds)
  assert isinstance(result, benchmark.BenchmarkResult)

  # Works with other iterators
  result = benchmark.benchmark(range(10))
  assert isinstance(result, benchmark.BenchmarkResult)


def test_benchmark_empty():
  # len() == 0
  with pytest.raises(ValueError, match='Cannot benchmark dataset with 0 elem'):
    benchmark.benchmark([])

  # len() == 1
  result = benchmark.benchmark([1])
  assert isinstance(result, benchmark.BenchmarkResult)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.scripts.documentation.build_catalog."""

import pathlib

import tensorflow_datasets as tfds  # pylint: disable=unused-import
from tensorflow_datasets.scripts.documentation import build_catalog
from tensorflow_datasets.scripts.documentation import doc_utils


def test_build_catalog(tmp_path: pathlib.Path):
  """Tests that build_catalog correctly generates the index."""
  build_catalog.build_catalog(
      datasets=['mnist', 'coco'],
      catalog_dir=tmp_path,
      doc_util_paths=doc_utils.DocUtilPaths(
          fig_base_path=None,
          df_base_path=None,
          nightly_path=None,
      ),
      include_collections=False,
  )
  assert sorted(f.name for f in tmp_path.iterdir()) == [
      '_toc.yaml',
      'coco.md',
      'mnist.md',
      'overview.md',
  ]

  content = tmp_path.joinpath('_toc.yaml').read_text()
  assert 'coco' in content
  assert 'mnist' in content

  content = tmp_path.joinpath('overview.md').read_text()
  assert 'coco' in content
  assert 'mnist' in content


def test_build_catalog_with_collections(tmp_path: pathlib.Path):
  """Tests that build_catalog generates the index including collections."""
  build_catalog.build_catalog(
      datasets=['mnist', 'coco'],
      ds_collections=['longt5'],
      catalog_dir=tmp_path,
      doc_util_paths=doc_utils.DocUtilPaths(
          fig_base_path=None,
          df_base_path=None,
          nightly_path=None,
      ),
      include_collections=True,
  )
  assert sorted(f.name for f in tmp_path.iterdir()) == [
      '_toc.yaml',
      'coco.md',
      'longt5.md',
      'mnist.md',
      'overview.md',
  ]

  content = tmp_path.joinpath('_toc.yaml').read_text()
  assert 'coco' in content
  assert 'mnist' in content
  assert 'longt5' in content

  content = tmp_path.joinpath('overview.md').read_text()
  assert 'coco' in content
  assert 'mnist' in content
  assert 'longt5' in content



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test dataset builder without TFDV."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
import tensorflow_datasets as tfds

tf.compat.v1.enable_eager_execution()


class BuilderNoStatsTest(tfds.testing.TestCase):

  def test_no_tfdv_dir(self):
    with tfds.testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = tfds.testing.DummyMnist(data_dir=tmp_dir)
      builder.download_and_prepare()  # tfdv not present. Statistics skipped.

      # Statistics skipped
      num_examples = int(builder.info.splits['train'].statistics.num_examples)
      self.assertEqual(num_examples, 0)


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for partial_decode."""

import pytest

import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import features as features_lib


def _extract_features(feature, expected_feature):
  decoder = decode.PartialDecoding(expected_feature)
  return decoder.extract_features(feature)


def test_extract_features():
  features = features_lib.FeaturesDict({
      'img': features_lib.Image(shape=(256, 256, 3)),
      'img2': features_lib.Image(shape=(256, 256, 3)),
      'metadata': {
          'label': features_lib.ClassLabel(num_classes=4),
          'other': tf.string,
      },
      'sequence': features_lib.Sequence({
          'x': tf.int64,
          'y': tf.int64,
      }),
      'sequence_flat': features_lib.Sequence(tf.int64),
  })

  result = _extract_features(
      feature=features,
      expected_feature={},
  )
  testing.assert_features_equal(result, features_lib.FeaturesDict({}))

  # Feature spec accepted
  result = _extract_features(
      feature=features,
      expected_feature={
          'img': features_lib.Image(shape=(None, None, 3)),
          'metadata': {
              'other': tf.string,
          },
          'sequence': features_lib.Sequence(
              {
                  'x': tf.int64,
              }
          ),
      },
  )
  testing.assert_features_equal(
      result,
      features_lib.FeaturesDict({
          'img': features_lib.Image(shape=(256, 256, 3)),
          'metadata': {
              'other': tf.string,
          },
          'sequence': features_lib.Sequence(
              {
                  'x': tf.int64,
              }
          ),
      }),
  )

  # Failure mode:
  # * Structure not matching
  # * Type not matching
  # * Shape/dtype not matching
  # * Sequence values not matching (e.g. try bad dtype)

  with pytest.raises(ValueError, match="Missing expected feature 'unknown'"):
    _extract_features(
        feature=features,
        expected_feature={
            'sequence': features_lib.Sequence(
                {
                    'unknown': tf.bool,
                }
            )
        },
    )

  with pytest.raises(ValueError, match="Missing expected feature 'non_exista"):
    _extract_features(
        feature=features,
        expected_feature={
            'non_existant': features_lib.Image(shape=(None, None, 3)),
        },
    )

  with pytest.raises(TypeError, match='Expected: Tensor.*. Got: Image'):
    _extract_features(
        feature=features,
        expected_feature={
            'img': features_lib.Tensor(shape=(256, 256, 3), dtype=tf.uint8),
        },
    )

  with pytest.raises(ValueError, match='Expected: Image.*. Got: Image'):
    _extract_features(
        feature=features,
        expected_feature={
            'img': features_lib.Image(shape=(None, None, 1)),
        },
    )

  with pytest.raises(ValueError, match='Expected: Tensor.*. Got: Tensor'):
    _extract_features(
        feature=features,
        expected_feature={
            'sequence_flat': features_lib.Sequence(tf.float32),  # Wrong dtype
        },
    )


def test_extract_features_values():
  features = features_lib.FeaturesDict({
      'img': features_lib.Image(shape=(256, 256, 3)),
      'img2': features_lib.Image(shape=(256, 256, 3)),
      'metadata': {
          'label': features_lib.ClassLabel(num_classes=4),
          'other': tf.string,
      },
      'sequence': features_lib.Sequence({
          'x': tf.int64,
          'y': tf.int64,
      }),
      'sequence_flat': features_lib.Sequence(tf.int64),
  })

  result = _extract_features(
      feature=features,
      expected_feature={
          'img': True,
          'img2': False,
          'unknown_key': False,  # Extra keys are filtered
          'metadata': ['label'],
          'sequence': {'y'},
          'sequence_flat': True,
      },
  )
  testing.assert_features_equal(
      result,
      features_lib.FeaturesDict({
          'img': features_lib.Image(shape=(256, 256, 3)),
          'metadata': {
              'label': features_lib.ClassLabel(num_classes=4),
          },
          'sequence': features_lib.Sequence(
              {
                  'y': tf.int64,
              }
          ),
          'sequence_flat': features_lib.Sequence(tf.int64),
      }),
  )

  result = _extract_features(
      feature=features,
      expected_feature={'metadata', 'sequence'},
  )
  testing.assert_features_equal(
      result,
      features_lib.FeaturesDict({
          'metadata': {
              'label': features_lib.ClassLabel(num_classes=4),
              'other': tf.string,
          },
          'sequence': features_lib.Sequence({
              'x': tf.int64,
              'y': tf.int64,
          }),
      }),
  )

  # Test, mixing Features with non-features.
  result = _extract_features(
      feature=features,
      expected_feature={
          'img': features_lib.Image(),
          'sequence': {
              'x': tf.int64,
              'y': False,
          },
      },
  )
  testing.assert_features_equal(
      result,
      features_lib.FeaturesDict({
          'img': features_lib.Image(shape=(256, 256, 3)),
          'sequence': features_lib.Sequence(
              {
                  'x': tf.int64,
              }
          ),
      }),
  )


def test_partial_decode(dummy_mnist: testing.DummyMnist):
  ds = dummy_mnist.as_dataset(
      split='train',
      decoders=decode.PartialDecoding(
          {
              'image': features_lib.Image(shape=(None, None, 1)),
          }
      ),
  )
  assert ds.element_spec == {
      'image': tf.TensorSpec(shape=(28, 28, 1), dtype=tf.uint8)
  }
  for ex in ds.take(2):
    assert ex['image'].shape == (28, 28, 1)


def test_partial_decode_with_skip_decode(dummy_mnist: testing.DummyMnist):
  ds = dummy_mnist.as_dataset(
      split='train',
      decoders=decode.PartialDecoding(
          {
              'image': features_lib.Image(shape=(None, None, 1)),
          },
          decoders={
              'image': decode.SkipDecoding(),
          },
      ),
  )
  assert ds.element_spec == {'image': tf.TensorSpec(shape=(), dtype=tf.string)}
  for ex in ds.take(2):
    assert ex['image'].dtype == tf.string



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Format-specific dataset builders for CoNLL-like formatted data.

It contains a ConllBuilderConfig and a ConllDatasetBuilder which are used to
initialize TFDS datasets based on CoNLL-like formatted data.

Typical usage example:
class MyConllDataset(tfds.dataset_builders.ConllDatasetBuilder):
  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [conll_lib.CONLL_2003_CONFIG]

  def _info(self) -> tfds.core.DatasetInfo:
    ...

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    ...
"""
from typing import List, Optional, OrderedDict, Sequence, Union

from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features.features_dict import FeaturesDict


# TODO(b/241346210): Should update ConllBuilderConfig to @dataclasses.dataclass.
class ConllBuilderConfig(dataset_builder.BuilderConfig):
  """Base class for CoNLL-like formatted data configuration.

  Attributes:
    separator: The separator used for splitting feature columns in the input
      lines. For CoNLL-formatted data, this is usually a tab or a space.
    ordered_features: An OrderedDict specifying the features names and their
      type, in the same order as they appear as columns in the input lines.
  """

  def __init__(
      self,
      *,
      separator: str,
      ordered_features: OrderedDict[str, feature_lib.FeatureConnector],
      **kwargs,
  ):
    """Initializes the builder config for Conll datasets.

    Args:
      separator: The separator used for splitting feature columns in the input
        lines. For CoNLL-formatted data, this is usually a tab or a space.
      ordered_features: An OrderedDict specifying the features names and their
        type, in the same order as they appear as columns in the input lines.
      **kwargs: keyword arguments forwarded to super.
    """
    super(ConllBuilderConfig, self).__init__(**kwargs)
    self.separator = separator
    self.ordered_features = ordered_features

  @property
  def features_dict(self) -> FeaturesDict:
    return FeaturesDict(self.ordered_features)


class ConllDatasetBuilder(
    dataset_builder.GeneratorBasedBuilder, skip_registration=True
):
  """Base class for CoNLL-like formatted datasets.

  It provides functionalities to ease the processing of CoNLL-like datasets.
  Users can overwrite `_generate_examples` to customize the pipeline.
  """

  BUILDER_CONFIGS: Sequence[ConllBuilderConfig] = []

  @property
  def builder_config(self) -> ConllBuilderConfig:
    """`tfds.core.BuilderConfig` for this builder."""
    return self._builder_config  # pytype: disable=bad-return-type  # always-use-return-annotations

  def create_dataset_info(
      self,
      description: Optional[str] = None,
      supervised_keys: Optional[dataset_info.SupervisedKeysType] = None,
      homepage: Optional[str] = None,
      citation: Optional[str] = None,
  ) -> dataset_info.DatasetInfo:
    """Initializes `dataset_info.DatasetInfo` for Conll datasets.

    Args:
      description: [DEPRECATED] A short, markdown-formatted description of the
        dataset. Prefer placing description in `README.md` file.
      supervised_keys:  Specifies the input structure for supervised learning,
        if applicable for the dataset, used with "as_supervised". Typically this
        is a `(input_key, target_key)` tuple.
      homepage: The homepage of the dataset, if applicable for this dataset.
      citation: [DEPRECATED] The citation to use for this dataset, if applicable
        for this dataset. Prefer placing citations in `CITATIONS.bib` file.

    Returns:
      `dataset_info.DatasetInfo` for Conll datasets, populated with the values
      from the provided arguments.
    """
    return self.dataset_info_from_configs(
        description=description,
        features=self.builder_config.features_dict,
        supervised_keys=supervised_keys,
        homepage=homepage,
        citation=citation,
    )

  def _generate_examples(
      self,
      filepaths: Union[epath.PathLike, List[epath.PathLike]],
  ) -> split_builder_lib.SplitGenerator:
    """Processes CoNLL-like datasets and generate examples.

    Args:
      filepaths: The filepaths of the input data. Could be a list of paths for
        multiple input files, or a single path.

    Yields:
      Generated examples.
    """
    path = filepaths if isinstance(filepaths, list) else [filepaths]

    input_sequences = {feature: [] for feature in self.info.features}

    example_id = 0
    for filepath in path:
      with epath.Path(filepath).open() as f:
        for line in f:
          if line.startswith("-DOCSTART-") or line == "\n" or not line:
            if input_sequences["tokens"]:
              yield example_id, input_sequences
              example_id += 1
              input_sequences = {feature: [] for feature in self.info.features}
          else:
            splits = line.split(self.builder_config.separator)
            if len(splits) != len(self.builder_config.ordered_features):
              raise ValueError(
                  f"Mismatch in the number of features found in line: {line}\n"
                  f"Should be {len(self.builder_config.ordered_features)}, "
                  f"but found {len(splits)}"
              )
            for index, feature in enumerate(
                self.builder_config.ordered_features.keys()
            ):
              input_sequences[feature].append(splits[index].rstrip())

      # Last example from file.
      yield example_id, input_sequences

      # Initializing a new empty example in case of multiple filepaths in path.
      example_id += 1
      input_sequences = {feature: [] for feature in self.info.features}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
#  Copyright 2022 The TensorFlow Authors
#  Copyright 2022 The HuggingFace Dataset Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data for the XTREME-S Benchmark.

This corresponds to XTREME-S on HugginFace, defined at:
https://huggingface.co/datasets/google/xtreme_s/blob/main/xtreme_s.py
"""

import collections
import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@article{conneau2022xtreme,
  title={XTREME-S: Evaluating Cross-lingual Speech Representations},
  author={Conneau, Alexis and Bapna, Ankur and Zhang, Yu and Ma, Min and von Platen, Patrick and Lozhkov, Anton and Cherry, Colin and Jia, Ye and Rivera, Clara and Kale, Mihir and others},
  journal={arXiv preprint arXiv:2203.10752},
  year={2022}
}
"""

_DESCRIPTION = """\
XTREME-S covers four task families: speech recognition, classification, speech-to-text translation and retrieval. Covering 102
languages from 10+ language families, 3 different domains and 4
task families, XTREME-S aims to simplify multilingual speech
representation evaluation, as well as catalyze research in “universal” speech representation learning.

In this version, only the FLEURS dataset is provided, which covers speech
recognition and speech-to-text translation.
"""

# Full English name to language subtag.
_FLEURS_LANG_TO_ID = collections.OrderedDict([
    ("Afrikaans", "af"),
    ("Amharic", "am"),
    ("Arabic", "ar"),
    ("Armenian", "hy"),
    ("Assamese", "as"),
    ("Asturian", "ast"),
    ("Azerbaijani", "az"),
    ("Belarusian", "be"),
    ("Bengali", "bn"),
    ("Bosnian", "bs"),
    ("Bulgarian", "bg"),
    ("Burmese", "my"),
    ("Catalan", "ca"),
    ("Cebuano", "ceb"),
    ("Mandarin Chinese", "cmn_hans"),
    ("Cantonese Chinese", "yue_hant"),
    ("Croatian", "hr"),
    ("Czech", "cs"),
    ("Danish", "da"),
    ("Dutch", "nl"),
    ("English", "en"),
    ("Estonian", "et"),
    ("Filipino", "fil"),
    ("Finnish", "fi"),
    ("French", "fr"),
    ("Fula", "ff"),
    ("Galician", "gl"),
    ("Ganda", "lg"),
    ("Georgian", "ka"),
    ("German", "de"),
    ("Greek", "el"),
    ("Gujarati", "gu"),
    ("Hausa", "ha"),
    ("Hebrew", "he"),
    ("Hindi", "hi"),
    ("Hungarian", "hu"),
    ("Icelandic", "is"),
    ("Igbo", "ig"),
    ("Indonesian", "id"),
    ("Irish", "ga"),
    ("Italian", "it"),
    ("Japanese", "ja"),
    ("Javanese", "jv"),
    ("Kabuverdianu", "kea"),
    ("Kamba", "kam"),
    ("Kannada", "kn"),
    ("Kazakh", "kk"),
    ("Khmer", "km"),
    ("Korean", "ko"),
    ("Kyrgyz", "ky"),
    ("Lao", "lo"),
    ("Latvian", "lv"),
    ("Lingala", "ln"),
    ("Lithuanian", "lt"),
    ("Luo", "luo"),
    ("Luxembourgish", "lb"),
    ("Macedonian", "mk"),
    ("Malay", "ms"),
    ("Malayalam", "ml"),
    ("Maltese", "mt"),
    ("Maori", "mi"),
    ("Marathi", "mr"),
    ("Mongolian", "mn"),
    ("Nepali", "ne"),
    ("Northern-Sotho", "nso"),
    ("Norwegian", "nb"),
    ("Nyanja", "ny"),
    ("Occitan", "oc"),
    ("Oriya", "or"),
    ("Oromo", "om"),
    ("Pashto", "ps"),
    ("Persian", "fa"),
    ("Polish", "pl"),
    ("Portuguese", "pt"),
    ("Punjabi", "pa"),
    ("Romanian", "ro"),
    ("Russian", "ru"),
    ("Serbian", "sr"),
    ("Shona", "sn"),
    ("Sindhi", "sd"),
    ("Slovak", "sk"),
    ("Slovenian", "sl"),
    ("Somali", "so"),
    ("Sorani-Kurdish", "ckb"),
    ("Spanish", "es"),
    ("Swahili", "sw"),
    ("Swedish", "sv"),
    ("Tajik", "tg"),
    ("Tamil", "ta"),
    ("Telugu", "te"),
    ("Thai", "th"),
    ("Turkish", "tr"),
    ("Ukrainian", "uk"),
    ("Umbundu", "umb"),
    ("Urdu", "ur"),
    ("Uzbek", "uz"),
    ("Vietnamese", "vi"),
    ("Welsh", "cy"),
    ("Wolof", "wo"),
    ("Xhosa", "xh"),
    ("Yoruba", "yo"),
    ("Zulu", "zu"),
])

# Short name (language subtag) to full English name.
_FLEURS_LANG_SHORT_TO_LONG = {v: k for k, v in _FLEURS_LANG_TO_ID.items()}

# Language codes, lowercase, underscore-separated versions of BCP-47 tags,
# including language and region subtags, and sometimes script subtags.
_FLEURS_LANG = sorted([
    "af_za",
    "am_et",
    "ar_eg",
    "as_in",
    "ast_es",
    "az_az",
    "be_by",
    "bn_in",
    "bs_ba",
    "ca_es",
    "ceb_ph",
    "cmn_hans_cn",
    "yue_hant_hk",
    "cs_cz",
    "cy_gb",
    "da_dk",
    "de_de",
    "el_gr",
    "en_us",
    "es_419",
    "et_ee",
    "fa_ir",
    "ff_sn",
    "fi_fi",
    "fil_ph",
    "fr_fr",
    "ga_ie",
    "gl_es",
    "gu_in",
    "ha_ng",
    "he_il",
    "hi_in",
    "hr_hr",
    "hu_hu",
    "hy_am",
    "id_id",
    "ig_ng",
    "is_is",
    "it_it",
    "ja_jp",
    "jv_id",
    "ka_ge",
    "kam_ke",
    "kea_cv",
    "kk_kz",
    "km_kh",
    "kn_in",
    "ko_kr",
    "ckb_iq",
    "ky_kg",
    "lb_lu",
    "lg_ug",
    "ln_cd",
    "lo_la",
    "lt_lt",
    "luo_ke",
    "lv_lv",
    "mi_nz",
    "mk_mk",
    "ml_in",
    "mn_mn",
    "mr_in",
    "ms_my",
    "mt_mt",
    "my_mm",
    "nb_no",
    "ne_np",
    "nl_nl",
    "nso_za",
    "ny_mw",
    "oc_fr",
    "om_et",
    "or_in",
    "pa_in",
    "pl_pl",
    "ps_af",
    "pt_br",
    "ro_ro",
    "ru_ru",
    "bg_bg",
    "sd_in",
    "sk_sk",
    "sl_si",
    "sn_zw",
    "so_so",
    "sr_rs",
    "sv_se",
    "sw_ke",
    "ta_in",
    "te_in",
    "tg_tj",
    "th_th",
    "tr_tr",
    "uk_ua",
    "umb_ao",
    "ur_pk",
    "uz_uz",
    "vi_vn",
    "wo_sn",
    "xh_za",
    "yo_ng",
    "zu_za",
])

# Full English name to language code.
_FLEURS_LONG_TO_LANG = {
    _FLEURS_LANG_SHORT_TO_LONG["_".join(k.split("_")[:-1]) or k]: k
    for k in _FLEURS_LANG
}

# Language code to full English name.
_FLEURS_LANG_TO_LONG = {v: k for k, v in _FLEURS_LONG_TO_LANG.items()}

_FLEURS_GROUP_TO_LONG = collections.OrderedDict({
    "western_european_we": [
        "Asturian",
        "Bosnian",
        "Catalan",
        "Croatian",
        "Danish",
        "Dutch",
        "English",
        "Finnish",
        "French",
        "Galician",
        "German",
        "Greek",
        "Hungarian",
        "Icelandic",
        "Irish",
        "Italian",
        "Kabuverdianu",
        "Luxembourgish",
        "Maltese",
        "Norwegian",
        "Occitan",
        "Portuguese",
        "Spanish",
        "Swedish",
        "Welsh",
    ],
    "eastern_european_ee": [
        "Armenian",
        "Belarusian",
        "Bulgarian",
        "Czech",
        "Estonian",
        "Georgian",
        "Latvian",
        "Lithuanian",
        "Macedonian",
        "Polish",
        "Romanian",
        "Russian",
        "Serbian",
        "Slovak",
        "Slovenian",
        "Ukrainian",
    ],
    "central_asia_middle_north_african_cmn": [
        "Arabic",
        "Azerbaijani",
        "Hebrew",
        "Kazakh",
        "Kyrgyz",
        "Mongolian",
        "Pashto",
        "Persian",
        "Sorani-Kurdish",
        "Tajik",
        "Turkish",
        "Uzbek",
    ],
    "sub_saharan_african_ssa": [
        "Afrikaans",
        "Amharic",
        "Fula",
        "Ganda",
        "Hausa",
        "Igbo",
        "Kamba",
        "Lingala",
        "Luo",
        "Northern-Sotho",
        "Nyanja",
        "Oromo",
        "Shona",
        "Somali",
        "Swahili",
        "Umbundu",
        "Wolof",
        "Xhosa",
        "Yoruba",
        "Zulu",
    ],
    "south_asian_sa": [
        "Assamese",
        "Bengali",
        "Gujarati",
        "Hindi",
        "Kannada",
        "Malayalam",
        "Marathi",
        "Nepali",
        "Oriya",
        "Punjabi",
        "Sindhi",
        "Tamil",
        "Telugu",
        "Urdu",
    ],
    "south_east_asian_sea": [
        "Burmese",
        "Cebuano",
        "Filipino",
        "Indonesian",
        "Javanese",
        "Khmer",
        "Lao",
        "Malay",
        "Maori",
        "Thai",
        "Vietnamese",
    ],
    "chinese_japanase_korean_cjk": [
        "Mandarin Chinese",
        "Cantonese Chinese",
        "Japanese",
        "Korean",
    ],
})

# Full English name to group name.
_FLEURS_LONG_TO_GROUP = {  # pylint: disable=g-complex-comprehension
    a: k for k, v in _FLEURS_GROUP_TO_LONG.items() for a in v
}

# Language code to group name.
_FLEURS_LANG_TO_GROUP = {
    _FLEURS_LONG_TO_LANG[k]: v for k, v in _FLEURS_LONG_TO_GROUP.items()
}

_ALL_DATASET_CONFIGS = {
    "fleurs": _FLEURS_LANG,
}

_ALL_CONFIGS = []  # e.g. mls.en, covost.en.sv, ...
for sub_data, langs in _ALL_DATASET_CONFIGS.items():
  for lang in langs:
    _ALL_CONFIGS.append(f"{sub_data}.{lang}")

# add "all" for all datasets except 'BABEL'
_ALL_CONFIGS += [
    "fleurs.all",
]

_DESCRIPTIONS = {
    "fleurs": (
        "FLEURS is the speech version of the FLORES machine translation"
        " benchmark, covering 2000 n-way parallel sentences in n=102"
        " languages."
    ),
}

_CITATIONS = {
    "fleurs": """\
@article{fleurs2022arxiv,
  title = {FLEURS: Few-shot Learning Evaluation of Universal Representations of Speech},
  author = {Conneau, Alexis and Ma, Min and Khanuja, Simran and Zhang, Yu and Axelrod, Vera and Dalmia, Siddharth and Riesa, Jason and Rivera, Clara and Bapna, Ankur},
  journal={arXiv preprint arXiv:2205.12446},
  url = {https://arxiv.org/abs/2205.12446},
  year = {2022},
}""",
}

_HOMEPAGE_URLS = {
    "fleurs": "https://arxiv.org/abs/2205.12446",
}

_DATA_URLS = {
    "fleurs": [
        "https://storage.googleapis.com/xtreme_translations/FLEURS102/{}.tar.gz"
    ],
}


class XtremeSConfig(tfds.core.BuilderConfig):
  """BuilderConfig for XTREME-S."""

  def __init__(
      self,
      name,
      dataset_name,
      lang_name,
      description,
      citation,
      homepage,
      data_urls,
  ):
    super().__init__(
        name=name,
        version=tfds.core.Version("2.0.0"),
        release_notes={
            "2.0.0": (
                "Initial release on TFDS, FLEURS-only. Named to match version"
                " 2.0.0 on huggingface which has the same FLEURS data ("
                " https://huggingface.co/datasets/google/xtreme_s)."
            )
        },
        description=self.description,
    )
    self.name = name
    self.dataset_name = dataset_name
    self.lang_name = lang_name
    self.description = description
    self.citation = citation
    self.homepage = homepage
    self.data_urls = data_urls


def _build_config(name):
  dataset_name = name.split(".")[0]
  lang_name = ".".join(name.split(".")[1:])

  return XtremeSConfig(
      name=name,
      dataset_name=dataset_name,
      lang_name=lang_name,
      description=_DESCRIPTIONS[dataset_name],
      citation=_CITATIONS[dataset_name],
      homepage=_HOMEPAGE_URLS[dataset_name],
      data_urls=_DATA_URLS[dataset_name],
  )


class XtremeS(tfds.core.GeneratorBasedBuilder):
  """XTREME-S Benchmark datasets."""

  DEFAULT_WRITER_BATCH_SIZE = 1000
  BUILDER_CONFIGS = [_build_config(name) for name in _ALL_CONFIGS]

  def _info(self):
    languages = _ALL_DATASET_CONFIGS[self.builder_config.dataset_name]

    if self.builder_config.dataset_name == "fleurs":
      features = tfds.features.FeaturesDict({
          "id": tfds.features.Scalar(
              tf.dtypes.int32,
              doc=(
                  "Source text identifier, consistent across all languages"
                  " to keep n-way parallelism of translations. Since each"
                  " transcription may be spoken by multiple speakers,"
                  " within each language multiple examples will also share"
                  " the same id."
              ),
          ),
          "num_samples": tfds.features.Scalar(
              tf.dtypes.int32, doc="Total number of frames in the audio"
          ),
          "path": tf.dtypes.string,
          "audio": tfds.features.Audio(sample_rate=16_000),
          "transcription": tfds.features.Text(doc="Normalized transcription."),
          "raw_transcription": tfds.features.Text(
              doc="Raw Transcription from FLoRes."
          ),
          "gender": tfds.features.ClassLabel(names=["male", "female", "other"]),
          "lang_id": tfds.features.ClassLabel(names=languages),
          "language": tfds.features.Text(
              doc=(
                  "Language encoded as lowercase, underscore-separated"
                  "version of a BCP-47 tag."
              )
          ),
          "lang_group_id": tfds.features.ClassLabel(
              names=list(_FLEURS_GROUP_TO_LONG.keys())
          ),
      })

      return tfds.core.DatasetInfo(
          builder=self,
          description=self.builder_config.description + "\n" + _DESCRIPTION,
          features=features,
          supervised_keys=("audio", "transcription"),
          homepage=self.builder_config.homepage,
          citation=self.builder_config.citation + "\n" + _CITATION,
      )

  def _split_generators(self, *args, **kwargs):
    if self.builder_config.dataset_name == "fleurs":
      return self._fleurs_split_generators(*args, **kwargs)

  def _generate_examples(self, *args, **kwargs):
    if self.builder_config.dataset_name == "fleurs":
      yield from self._fleurs_generate_examples(*args, **kwargs)

  def _fleurs_split_generators(self, dl_manager):
    data_url_format = self.builder_config.data_urls[0]

    if self.builder_config.lang_name == "all":
      data_urls = {l: data_url_format.format(l) for l in _FLEURS_LANG}
    else:
      data_urls = {
          self.builder_config.lang_name: data_url_format.format(
              self.builder_config.lang_name
          )
      }

    archive_path = dl_manager.download_and_extract(data_urls)
    audio_paths = {
        l: os.path.join(v, l, "audio") for l, v in archive_path.items()
    }
    text_paths = {l: os.path.join(v, l) for l, v in archive_path.items()}

    def create_split_generator(
        split: str, file_prefix: str
    ) -> tfds.core.SplitGenerator:
      return tfds.core.SplitGenerator(
          name=split,
          gen_kwargs={
              "audio_path": {
                  l: os.path.join(v, file_prefix)
                  for l, v in audio_paths.items()
              },
              "text_path": {
                  l: os.path.join(v, f"{file_prefix}.tsv")
                  for l, v in text_paths.items()
              },
          },
      )

    return [
        create_split_generator(
            split=tfds.core.Split.TRAIN, file_prefix="train"
        ),
        create_split_generator(
            split=tfds.core.Split.VALIDATION, file_prefix="dev"
        ),
        create_split_generator(split=tfds.core.Split.TEST, file_prefix="test"),
    ]

  def _fleurs_generate_examples(self, audio_path, text_path):
    key = 0

    gender_to_id = {"MALE": 0, "FEMALE": 1, "OTHER": 2}

    for lang_id in text_path.keys():
      text_file = text_path[lang_id]
      audio_dir = audio_path[lang_id]

      with tf.io.gfile.GFile(text_file, "r") as f:
        lines = f.readlines()
        for line in lines:
          (
              index,
              file_name,
              raw_transcription,
              transcription,
              _,  # character transcription
              num_samples,
              gender,
          ) = line.strip().split("\t")

          lang_group = _FLEURS_LANG_TO_GROUP[lang_id]

          yield key, {
              "id": int(index),
              "path": os.path.join(audio_dir, file_name),
              "audio": os.path.join(audio_dir, file_name),
              "raw_transcription": raw_transcription,
              "transcription": transcription,
              "num_samples": int(num_samples),
              "gender": gender_to_id[gender],
              "lang_id": _FLEURS_LANG.index(lang_id),
              "language": _FLEURS_LANG_TO_LONG[lang_id],
              "lang_group_id": list(_FLEURS_GROUP_TO_LONG.keys()).index(
                  lang_group
              ),
          }
          key += 1



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.image_feature."""

import numpy as np
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import features as features_lib


class LabeledImageFeatureTest(testing.FeatureExpectationsTestCase):

  def test_images(self):
    rng = np.random.default_rng()
    img = rng.integers(256, size=(28, 28, 1), dtype=np.uint8)

    self.assertFeature(
        feature=features_lib.LabeledImage(
            labels=['background', 'car', 'truck'],
            shape=(28, 28, 1),
        ),
        shape=(28, 28, 1),
        dtype=tf.uint8,
        tests=[
            # Numpy array
            testing.FeatureExpectationItem(
                value=img,
                expected=img,
            ),
        ],
        test_attributes=dict(
            num_classes=3,
            names=['background', 'car', 'truck'],
            _use_colormap=True,
        ),
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for error_utils."""
import pytest

from tensorflow_datasets.core.utils import error_utils


def test_default_empty_context_stack():
  with pytest.raises(ValueError, match=''):
    with error_utils.reraise_with_context(ValueError):
      raise ValueError


def test_raise_with_context():
  with pytest.raises(ValueError, match='Adding context'):
    with error_utils.reraise_with_context(ValueError):
      error_utils.add_context('Adding context')
      raise ValueError


def test_raise_with_multiple_contexts():
  with pytest.raises(
      ValueError, match='\n'.join(['Adding context', 'Adding context 2'])
  ):
    with error_utils.reraise_with_context(ValueError):
      error_utils.add_context('Adding context')
      error_utils.add_context('Adding context 2')
      raise ValueError


def test_do_not_raise_error_if_nested_reraise_with_context():
  with pytest.raises(
      ValueError, match='\n'.join(['Adding context', 'Adding context 2'])
  ):
    with error_utils.reraise_with_context(ValueError):
      error_utils.add_context('Adding context')
      with error_utils.reraise_with_context(ValueError):
        error_utils.add_context('Adding context 2')
        raise ValueError


def test_add_context_outside_contextmanager():
  with pytest.raises(
      AttributeError,
      match=(
          'add_context called outside of reraise_with_context contextmanager.'
      ),
  ):
    error_utils.add_context('Adding context')
